# #!/usr/bin/python3
# -*- coding: utf-8 -*-


import itertools as itt
import datetime as dtm


import numpy as np
import scipy.optimize as scpopt


# local
from . import _class02_x0_scale_bounds as _x0_scale_bounds


#############################################
#############################################
#       DEFAULTS
#############################################


#############################################
#############################################
#       Main
#############################################


def main(
    coll=None,
    # keys
    key=None,
    key_model=None,
    key_data=None,
    key_lamb=None,
    # covarance
    ref_cov=None,
    shape_cov=None,
    # lamb, data, axis
    lamb=None,
    data=None,
    sigma=None,
    axis=None,
    ravel=None,
    # binning
    binning=None,
    # options
    chain=None,
    dscales=None,
    dbounds_low=None,
    dbounds_up=None,
    dx0=None,
    # solver options
    solver=None,
    dsolver_options=None,
    absolute_sigma=None,
    # options
    strict=None,
    verb=None,
    verb_scp=None,
    timing=None,
):
    """ Fit 1d spectra

    """

    # -----------
    # verb

    if verb >= 2:
        msg = (
            "Preparing input data (scales, bounds, x0...)\n"
        )
        print(msg)

    # ------------
    # iok
    # ------------

    # iok_all
    iok_all, iok_reduced = get_iok_all(
        coll=coll,
        key=key,
        axis=axis,
    )

    # trivial
    if not np.any(iok_reduced):
        return

    # ravel
    if ravel is True:
        iok_all = iok_all[:, None]
        iok_reduced = np.array([iok_reduced])

    # ------------
    # lamb, data, dind
    # ------------

    dind = coll.get_spectral_model_variables_dind(key_model)
    lk_xfree = coll.get_spectral_model_variables(key_model, 'free')['free']

    wsm = coll._which_model
    dmodel = coll.dobj[wsm][key_model]['dmodel']

    # -----------------
    # optionnal binning
    # -----------------

    dbinning = coll.get_spectral_fit_binning_dict(
        binning=binning,
        lamb=lamb,
        iok=iok_all,
        axis=axis,
    )

    # ------------
    # get scale, bounds
    # ------------

    # check dscales
    dscales = _x0_scale_bounds._get_dict(
        lx_free_keys=lk_xfree,
        dmodel=dmodel,
        din=dscales,
        din_name='dscales',
    )

    # check dbounds_low
    dbounds_low = _x0_scale_bounds._get_dict(
        lx_free_keys=lk_xfree,
        dmodel=dmodel,
        din=dbounds_low,
        din_name='dbounds_low',
    )

    # check dbounds_up
    dbounds_up = _x0_scale_bounds._get_dict(
        lx_free_keys=lk_xfree,
        dmodel=dmodel,
        din=dbounds_up,
        din_name='dbounds_up',
    )

    # get scales and bounds
    scales, bounds0, bounds1 = _x0_scale_bounds._get_scales_bounds(
        nxfree=len(lk_xfree),
        lamb=lamb,
        data=data,
        iok=iok_all,
        axis=axis,
        dind=dind,
        dscales=dscales,
        dbounds_low=dbounds_low,
        dbounds_up=dbounds_up,
        # binning
        dbinning=dbinning,
    )

    # ------------
    # get x0
    # ------------

    # check dx0
    dx0 = _x0_scale_bounds._get_dict(
        lx_free_keys=lk_xfree,
        dmodel=dmodel,
        din=dx0,
        din_name='dx0',
    )

    # ------------
    # get functions
    # ------------

    # func_cost, func_jac
    dfunc = coll.get_spectral_fit_func(
        key=key_model,
        func=['sum', 'cost', 'jac'],
    )

    # ------------
    # Main loop
    # ------------

    dout = _loop(
        coll=coll,
        key=key,
        # lamb, data, axis
        lamb=lamb,
        data=data,
        sigma=sigma,
        axis=axis,
        # covarance
        ref_cov=ref_cov,
        shape_cov=shape_cov,
        # iok
        dind=dind,
        iok_all=iok_all,
        iok_reduced=iok_reduced,
        # x0, bounds, scale
        scales=scales,
        bounds0=bounds0,
        bounds1=bounds1,
        dx0=dx0,
        # options
        chain=chain,
        dbinning=dbinning,
        # func
        func_sum=dfunc.get('sum'),
        func_cost=dfunc.get('cost'),
        func_jac=dfunc['jac'],
        # solver options
        solver=solver,
        dsolver_options=dsolver_options,
        absolute_sigma=absolute_sigma,
        # options
        lk_xfree=lk_xfree,
        strict=strict,
        verb=verb,
        verb_scp=verb_scp,
        timing=timing,
    )

    return dout


#############################################
#############################################
#       get iok
#############################################


def get_iok_all(
    coll=None,
    key=None,
    axis=None,
):

    # ----------
    # prepare
    # ----------

    wsf = coll._which_fit
    kiok = coll.dobj[wsf][key]['dvalid']['iok']
    iok = coll.ddata[kiok]['data']
    meaning = coll.dobj[wsf][key]['dvalid']['meaning']

    # ----------
    # iok_all
    # ----------

    # get list of valid indices
    lind_valid = [
        k0 for k0, v0 in meaning.items()
        if any([ss in v0 for ss in ['ok', 'incl']])
    ]

    # iok_all
    iok_all = (iok == int(lind_valid[0]))
    for k0 in lind_valid[1:]:
        iok_all = np.logical_or(iok_all, (iok == int(k0)))

    iok_reduced = np.any(iok_all, axis=axis)

    return iok_all, iok_reduced


#############################################
#############################################
#       Main loop
#############################################


def _loop(
    coll=None,
    key=None,
    # lamb, data, axis
    lamb=None,
    data=None,
    sigma=None,
    axis=None,
    # covarance
    ref_cov=None,
    shape_cov=None,
    # iok
    dind=None,
    iok_all=None,
    iok_reduced=None,
    # x0, bounds, scale
    scales=None,
    bounds0=None,
    bounds1=None,
    dx0=None,
    # options
    chain=None,
    dbinning=None,
    # func
    func_sum=None,
    func_cost=None,
    func_jac=None,
    # solver options
    solver=None,
    dsolver_options=None,
    absolute_sigma=None,
    # options
    lk_xfree=None,
    strict=None,
    verb=None,
    verb_scp=None,
    timing=None,
):

    # -----------------
    # prepare
    # -----------------

    # shape_reduced
    shape_reduced = tuple([
        ss for ii, ss in enumerate(data.shape)
        if ii != axis
    ])

    # shape_sol
    nxfree = len(lk_xfree)
    shape_sol = list(data.shape)
    shape_sol[axis] = nxfree

    # lind
    lind = [range(ss) for ss in shape_reduced]

    # timing init
    if timing is True:
        t0 = dtm.datetime.now()

    # -----------------
    # prepare verb
    # -----------------

    # verb init
    if verb == 1:
        end = '\r'
    elif verb in [2, 3]:
        end = '\n'

    # iterations
    if verb >= 1:
        sep = '  '
        ditems = {
            'spectrum ind': {
                'just': max(13, len(str(shape_reduced))*2+3),
                'val': None,
            },
            'nfev': {'just': 6, 'val': None},
            'cost_final': {'just': 10, 'val': None},
            'status': {'just': 6, 'val': None},
            # 'termination': {'just': 12, 'val': None},
        }
        litems = ['spectrum ind', 'nfev', 'cost_final', 'status']

        if verb in [1, 2]:
            msg = (
                sep.join([k0.ljust(ditems[k0]['just']) for k0 in litems])
                + '\n'
                + sep.join(['-'*ditems[k0]['just'] for k0 in litems])
            )
            print(msg)

    # -----------------
    # initialize
    # -----------------

    validity = np.zeros(shape_reduced, dtype=int)
    status = np.full(shape_reduced, np.nan)
    cost = np.full(shape_reduced, np.nan)
    chi2n = np.full(shape_reduced, np.nan)
    nfev = np.full(shape_reduced, np.nan)
    time = np.full(shape_reduced, np.nan)
    sol = np.full(shape_sol, np.nan)

    # covariance matrix
    if solver == 'scipy.curve_fit':
        cov = np.full(shape_cov, np.nan)
        sli_cov = [0 for ii in shape_cov]
        sli_cov[axis] = slice(None)
        sli_cov[axis+1] = slice(None)
        sli_cov = np.array(sli_cov)
        ind_cov = np.array(
            [ii for ii in range(len(shape_cov)) if ii not in [axis, axis+1]],
            dtype=int,
        )
        scales_cov = scales[:, None] * scales[None, :]
    else:
        cov = None

    message = ['' for ii in range(np.prod(shape_reduced))]
    errmsg = ['' for ii in range(np.prod(shape_reduced))]

    # ----------
    # slice_sol

    sli_sol = np.array([
        slice(None) if ii == axis else 0
        for ii, ss in enumerate(data.shape)
    ])
    ind_ind = np.array([ii for ii in range(data.ndim) if ii != axis])

    # -----------------
    # initialize parameter dict
    # -----------------

    dparams = {
        'scales': scales,
        'iok': None,
        'bin_ind': False if dbinning is False else dbinning['ind'],
        'bin_dlamb': None if dbinning is False else dbinning['dlamb'],
    }
    if solver == 'scipy.least_squares':
        dparams['lamb'] = lamb if dbinning is False else dbinning['lamb']
        dparams['data'] = None
    else:
        pass

    # -----------------
    # main loop
    # -----------------

    for ii, ind in enumerate(itt.product(*lind)):

        # -------------
        # check iok_all

        if not iok_reduced[ind]:
            message[ii] = 'no valid data'
            validity[ind] = -1
            continue

        # -------
        # slices

        sli_sol[ind_ind] = ind
        slii = tuple(sli_sol)

        # covariance
        if cov is not None:
            sli_cov[ind_cov] = ind

        # ---------------
        # x0

        # get x0
        if chain is False or ii == 0:
            x0 = _x0_scale_bounds._get_x0(
                nxfree=nxfree,
                lamb=lamb,
                data=data[slii],
                iok=iok_all[slii],
                axis=axis,
                dind=dind,
                dx0=dx0,
                scales=scales,
                dbinning=dbinning,
            )

        # ------
        # verb

        if verb == 3:
            msg = f"\nspectrum {ind} / {shape_reduced}"
            print(msg)

        # -----------
        # parameters

        dparams['iok'] = iok_all[slii]
        if dbinning is False:
            dparams['bin_iok'] = iok_all[slii]
        else:
            dparams['bin_iok'] = dbinning['iok'][slii]

        # -----------
        # try solving

        try:
            dti = None
            t0i = dtm.datetime.now()     # DB

            if solver == 'scipy.least_squares':

                # update data
                dparams['data'] = data[slii]

                # optimization
                res = scpopt.least_squares(
                    func_cost,
                    x0,
                    jac=func_jac,
                    bounds=(bounds0, bounds1),
                    x_scale='jac',
                    f_scale=1.0,
                    jac_sparsity=None,
                    args=(),
                    kwargs=dparams,
                    **dsolver_options,
                )
                dti = (dtm.datetime.now() - t0i).total_seconds()

                # time
                time[ind] = round(dti, ndigits=6)

                # x0 if chain
                if chain is True:
                    x0 = res.x

                # other outputs
                status[ind] = res.status
                cost[ind] = res.cost
                nfev[ind] = res.nfev

                chi2n[ind] = np.sqrt(cost[ind] * 2) / iok_all[slii].sum()

                # store scaled solution
                sol[slii] = res.x * scales
                # sol_x[ii, ~indx] = const[ii, :] / scales[ii, ~indx]

                # message
                message[ii] = res.message
                errmsg[ii] = ''

            elif solver == 'scipy.curve_fit':

                def func_sum2(xdata, *xfree, dparams=dparams):
                    return func_sum(np.r_[xfree], lamb=xdata, **dparams)

                def func_jac2(xdata, *xfree, dparams=dparams):
                    return func_jac(np.r_[xfree], lamb=xdata, **dparams)

                popt, pcov, infodict, mesg, ier = scpopt.curve_fit(
                    func_sum2,
                    lamb if dbinning is False else dbinning['lamb'],
                    data[slii][iok_all[slii]],
                    p0=x0,
                    sigma=sigma[slii][iok_all[slii]],
                    absolute_sigma=absolute_sigma,
                    check_finite=True,    # to be updated
                    bounds=(bounds0, bounds1),
                    jac=func_jac2,
                    **dsolver_options,
                )
                dti = (dtm.datetime.now() - t0i).total_seconds()

                # time
                time[ind] = round(dti, ndigits=6)

                # x0 if chain
                if chain is True:
                    x0 = popt

                # other outputs
                status[ind] = ier
                cost[ind] = np.sum(infodict['fvec']**2)
                nfev[ind] = infodict['nfev']

                chi2n[ind] = np.sqrt(cost[ind]) / iok_all[slii].sum()

                # store scaled solution
                sol[slii] = popt * scales
                cov[tuple(sli_cov)] = pcov * scales_cov

                # message
                message[ii] = mesg
                errmsg[ii] = ''

        # ---------------
        # manage failures

        except Exception as err:

            msg = (
                f"\nError for spect_fit '{key}' with solver = '{solver}':\n"
                + str(err)
            )
            lerr = [
                'is infeasible',
                'Each lower bound must be strictly less than',
            ]
            if any([ee in msg for ee in lerr]):
                msg += _add_err_bounds(
                    key=key,
                    lk_xfree=lk_xfree,
                    scales=scales,
                    x0=x0,
                    bounds0=bounds0,
                    bounds1=bounds1,
                )

            if strict:
                raise Exception(msg) from err
            else:
                errmsg[ii] = msg
                validity[ii] = -2

        # -------------
        # verb

        finally:

            if verb in [1, 2]:
                ditems['spectrum ind']['val'] = f"{ind} / {shape_reduced}"
                ditems['status']['val'] = f"{status[ind]:.0f}"
                ditems['cost_final']['val'] = f"{cost[ind]:.4e}"
                ditems['nfev']['val'] = f"{nfev[ind]:.0f}"

                msg = (
                    sep.join([
                        ditems[k0]['val'].ljust(ditems[k0]['just'])
                        for k0 in litems
                    ])
                )
                print(msg, end=end)

    # -------------
    # adjust verb
    # -------------

    if verb == 2:
        print()

    # --------------
    # prepare output
    # --------------

    dout = {
        'validity': validity,
        'sol': sol,
        # covariance
        'cov': cov,
        'ref_cov': ref_cov,
        # output
        'msg': np.reshape(message, shape_reduced),
        'nfev': nfev,
        'cost': cost,
        'chi2n': chi2n,
        'status': status,
        'time': time,
        'errmsg': np.reshape(errmsg, shape_reduced),
        'scales': scales,
        'bounds0': bounds0,
        'bounds1': bounds1,
        'x0': x0,
        # binning:
        'dbinning': dbinning,
        # solver
        'solver': solver,
        'dsolver_options': dsolver_options,
    }

    return dout


#############################################
#############################################
#       Errors
#############################################


def _add_err_bounds(
    key=None,
    lk_xfree=None,
    scales=None,
    x0=None,
    bounds0=None,
    bounds1=None,
):

    # -------------
    # is_out
    # -------------

    is_out = np.nonzero(
        (x0 < bounds0) | (x0 > bounds1) | (bounds0 > bounds1)
    )[0]

    # -------------
    # dout, din
    # -------------

    dout = {
        k0: {
            'scale': f"{scales[ii]:.3e}",
            'x0': f"{x0[ii]:.3e}",
            'bounds0': f"{bounds0[ii]:.3e}",
            'bounds1': f"{bounds1[ii]:.3e}",
        }
        for ii, k0 in enumerate(lk_xfree) if ii in is_out
    }

    # -------------
    # msg arrays
    # -------------

    head = ['var', 'scale', 'x0', 'bounds0', 'bounds1']

    arr_out = [
        (k0, v0['scale'], v0['x0'], v0['bounds0'], v0['bounds1'])
        for k0, v0 in dout.items()
    ]

    arr_in = [
        (k0, v0['scale'], v0['x0'], v0['bounds0'], v0['bounds1'])
        for k0, v0 in dout.items()
    ]

    # max_just
    max_just = np.max([
        [len(ss) for ss in head]
        + [np.max([len(ss) for ss in arr]) for arr in arr_out]
        + [np.max([len(ss) for ss in arr]) for arr in arr_in]
    ])

    # -------------
    # msg
    # -------------

    lstr = [
        " ".join([ss.ljust(max_just) for ss in head]),
        " ".join(['-'*max_just for ss in head]),
    ]
    for arr in arr_out:
        lstr.append(" ".join([ss.ljust(max_just) for ss in arr]))

    msg = f" for spectral_fit '{key}'\n\n" + "\n".join(lstr)

    return msg


# -*- coding: utf-8 -*-
"""
Created on Tue Feb 20 17:33:22 2024

@author: dvezinet
"""


import os
import copy


import numpy as np
import datastock as ds


_DINDOK = {
    '0': 'ok',
    '-1': 'mask',
    '-2': 'out of domain',
    '-3': 'neg, inf or NaN',
    '-4': 'S/N valid, excl. (fract.)',
    '-5': 'S/N non-valid, incl. (fract.)',
    '-6': 'S/N non-valid, excl. (fract.)',
    '-7': 'S/N valid, excl. (bs)',
    '-8': 'S/N non-valid, incl. (bs)',
}


#############################################
#############################################
#       main
#############################################


def mask_domain(
    # resources
    coll=None,
    key_data=None,
    key_lamb=None,
    key_bs_vect=None,
    # options
    dvalid=None,
    ref=None,
    ref0=None,
    shape0=None,
):

    # ------------
    # check dvalid
    # ------------

    dvalid = _check_dvalid(dvalid)

    # ------------
    # check domain
    # ------------

    domain = _check_domain(
        domain=dvalid.get('domain'),
        key_lamb=key_lamb,
        key_bs_vect=key_bs_vect,
    )

    # ----------
    # Apply
    # ----------

    lk = [k0 for k0 in [key_lamb, key_bs_vect] if k0 is not None]
    ind = np.zeros(shape0, dtype=bool)
    for k0 in lk:

        # initialize
        c0 = (
            len(domain[k0]['spec']) == 1
            or all([isinstance(v1, tuple) for v1 in domain[k0]['spec']])
        )
        if c0 is True:
            indin = np.ones(coll.ddata[k0]['data'].shape, dtype=bool)
        else:
            indin = np.zeros(coll.ddata[k0]['data'].shape, dtype=bool)
        indout = np.zeros(coll.ddata[k0]['data'].shape, dtype=bool)

        # loop on domain bits
        for v1 in domain[k0]['spec']:

            indi = (
                (coll.ddata[k0]['data'] >= v1[0])
                & (coll.ddata[k0]['data'] <= v1[1])
            )

            # tuple => excluded
            if isinstance(v1, tuple):
                indout |= indi

            # list => included
            else:
                indin |= indi

        # store
        indi = (indin & (~indout))
        domain[k0]['ind'] = indi

        # apply to ind
        sli = tuple([
            indi if ii == ref0.index(coll.ddata[k0]['ref'][0]) else slice(None)
            for ii in range(len(ref0))
        ])
        ind[sli] = True

    # ---------------
    # check mask
    # ---------------

    dmask = _check_mask(
        coll=coll,
        mask=dvalid.get('mask'),
        ref0=ref0,
        shape0=shape0,
    )

    # -----------------
    # initialize iok
    # -----------------

    iok = np.zeros(coll.ddata[key_data]['data'].shape, dtype=int)

    # mask
    if dmask['key'] is not None:
        iout = ~coll.ddata[dmask['key']]['data']
        sli = tuple([
            iout
            if rr == ref0[0]
            else slice(None)
            for rr in ref
            if rr not in ref0[1:]
        ])
        iok[sli] = -1

    # domain
    iout = (~ind)
    if dmask['key'] is not None:
        iout &= coll.ddata[dmask['key']]['data']
    sli = tuple([
        iout
        if rr == ref0[0]
        else slice(None)
        for rr in ref
        if rr not in ref0[1:]
    ])
    iok[sli] = -2

    # -----------------
    # store in dvalid
    # -----------------

    if key_bs_vect is None:
        lkm = [k0 for k0, v0 in _DINDOK.items() if '(bs)' not in v0]
    else:
        lkm = list(_DINDOK.keys())

    inds = np.argsort([int(ss) for ss in lkm])[::-1]
    lindok = [lkm[ii] for ii in inds]

    dvalid.update({
        'domain': domain,
        'mask': dmask,
        'iok': iok,
        'meaning': {k0: _DINDOK[k0] for k0 in lindok},
    })

    return dvalid


#############################################
#############################################
#       check dvalid
#############################################


def _check_dvalid(dvalid=None):

    # ------------
    # None
    # ------------

    if dvalid is None:
        dvalid = {}

    # ------------
    # check
    # ------------

    c0 = (
        isinstance(dvalid, dict)
    )
    if not c0:
        msg = (
            "dvalid must be a dict with (optional) keys:\n"
        )
        raise Exception(msg)

    return dvalid


#############################################
#############################################
#       check domain
#############################################


def _check_domain(
    domain=None,
    key_lamb=None,
    key_bs_vect=None,
):

    # --------------
    # check domain
    # --------------

    # ------------
    # special case

    if key_bs_vect is None and not isinstance(domain, dict):
        domain = {key_lamb: domain}

    # --------
    # if None

    lk = [k0 for k0 in [key_lamb, key_bs_vect] if k0 is not None]
    if domain is None:
        domain = {
            k0: {
                'spec': [np.inf*np.r_[-1., 1.]],
                'minmax': np.inf*np.r_[-1., 1.],
            }
            for k0 in lk
        }

    # ----------
    # general

    c0 = (
        isinstance(domain, dict)
        and all([k0 in lk for k0 in domain.keys()])
    )
    if not c0:
        msg = (
            "Arg domain must be a dict with keys:\n"
            + "\n".join([f"\t- {k0}" for k0 in lk])
            + "\nProvided:\n{domain}\n"
        )
        raise Exception(msg)

    # deepcopy to avoid modifying in place
    domain = copy.deepcopy(domain)

    # ------------
    # loop on keys
    # ------------

    for k0 in lk:

        # -------
        # trivial

        if domain.get(k0) is None:
            domain[k0] = {
                'spec': [np.inf*np.r_[-1., 1.]],
                'minmax': np.inf*np.r_[-1., 1.],
            }
            continue

        # --------
        # sequence

        if not isinstance(domain[k0], dict):
            domain[k0] = {'spec': domain[k0]}
        else:
            domain = copy.deepcopy(domain)

        # --------
        # subdict

        c0 = (
            isinstance(domain[k0], dict)
            and all([k1 in ['spec', 'minmax'] for k1 in domain[k0].keys()])
        )
        if not c0:
            msg = (
                f"Arg domain['{k0}'] must be a dict with keys:\n"
                + "\n".join([f"\t- {k0}" for k0 in ['spec', 'minmax']])
                + f"\nProvided:\n{domain[k0]}\n"
            )
            raise Exception(msg)

        # -------------
        # check spec

        spec = domain[k0]['spec']
        if isinstance(spec, tuple):
            spec = [spec]

        c0 = (
            isinstance(spec, (list, np.ndarray))
            and len(spec) == 2
            and all([np.isscalar(ss) for ss in spec])
        )
        if c0:
            spec = [spec]

        c0 = (
            isinstance(spec, list)
            and all([
                isinstance(s0, (list, np.ndarray, tuple))
                and len(s0) == 2
                and all([np.isscalar(s1) for s1 in s0])
                and not np.any(np.isnan(s0))
                and s0[0] < s0[1]
                for s0 in spec
            ])
        )
        if not c0:
            msg = (
                f"Arg domain['{k0}']['spec'] must be a list of list/tuples\n"
                "\t Each must be a sequence of 2 increasing floats\n"
                f"Provided:\n{domain[k0]['spec']}"
            )
            raise Exception(msg)

        # -------
        # minmax

        domain[k0]['minmax'] = (
            np.nanmin(domain[k0]['spec']),
            np.nanmax(domain[k0]['spec']),
        )

    return domain


#############################################
#############################################
#       check mask
#############################################


def _mask_err(mask, ref=None, shape=None, lok=None):
    return Exception(
        "Arg mask must be either:\n"
        "\t- a str (path/file.ext) to a valid .npy file\n"
        f"\t- a key to a known data array with ref = {ref}\n"
        f"\t\tavailable: {lok}\n"
        f"\t- a np.narray of bool and of shape = {shape}\n"
        f"Provided:\n{mask}\n"
    )


def _check_mask(
    coll=None,
    mask=None,
    ref0=None,
    shape0=None,
):

    # ----------------
    # prepare
    # ----------------

    pfe = None
    key = None

    lok = [
        k0 for k0, v0 in coll.ddata.items()
        if v0['ref'] == ref0
        and v0['data'].dtype.name == 'bool'
    ]

    err = _mask_err(mask, ref=ref0, shape=shape0, lok=lok)


    # default key
    lout = [
        int(k0[4:]) for k0 in coll.ddata.keys()
        if k0.startswith('mask')
        and all([ss.isnumeric() for ss in k0[4:]])
    ]
    if len(lout) == 0:
        nmax = 0
    else:
        nmax = max(lout) + 1
    key0 = f"mask{nmax:02.0f}"

    # ----------------
    # str
    # ----------------

    if isinstance(mask, str):

        if os.path.isfile(mask) and mask.endswith('.npy'):
            pfe = str(mask)
            mask = np.load(pfe)
            key = key0

        elif mask in lok:
            key = mask

        else:
            raise err

    # --------------
    # numpy array
    # --------------

    if isinstance(mask, np.ndarray):

        c0 = (
            mask.shape == shape0
            and mask.dtype.name == 'bool'
        )
        if c0:
            key = key0

        else:
            raise err

    elif mask is None:
        pass

    else:
        raise err

    # ----------------------
    # store
    # ----------------------

    dmask = {
        'pfe': pfe,
        'key': key,
    }

    if key is not None and key not in coll.ddata.keys():
        coll.add_data(
            key=key,
            data=mask,
            ref=ref0,
            units=None,
            quant='bool',
            dim='mask',
        )

    return dmask


#############################################
#############################################
#       validity
#############################################


def valid(
    coll=None,
    key=None,
    key_data=None,
    key_lamb=None,
    key_bs=None,
    dvalid=None,
    ref=None,
    ref0=None,
):

    # -----------------
    # check inputs
    # -----------------

    # check nsigma, fraction, focus
    dvalid = _check_dvalid_valid(
        dvalid=dvalid,
        key_lamb=key_lamb,
    )

    # -----------------
    # prepare
    # -----------------

    wsf = coll._which_fit
    iok = dvalid['iok']
    data = coll.ddata[key_data]['data']
    lamb = coll.ddata[key_lamb]['data']

    # -----------------
    # nan, neg, inf
    # -----------------

    iokb = ((iok == 0) & (~np.isfinite(data)))
    if dvalid['positive'] is True:
        iokb &= (data >= 0)

    iok[iokb] = -3

    # update iokb
    iokb = (iok == 0)

    # safety check
    if not np.any(iokb):
        msg = (
            "Not a single valid data point for\n"
            f"\t- {wsf}: '{key}'\n"
            f"\t- key_data: '{key_data}'\n"
            f"\t- key_lamb: '{key_lamb}'\n"
            f"\t- mask: {dvalid['mask']['key']}\n"
            f"\t- domain: {dvalid['domain']}\n"
        )
        raise Exception(msg)

    # ----------------------------
    # Recompute domain min, max form valid data
    # ----------------------------

    if dvalid['update_domain'] is True:
        for k0, v0 in dvalid['domain'].items():
            refi = coll.ddata[k0]['ref'][0]
            axis = tuple([ii for ii, rr in enumerate(ref) if rr != refi])
            dvalid['domain'][k0]['minmax'] = [
                np.nanmin(coll.ddata[k0]['data'][np.any(iokb, axis=axis)]),
                np.nanmax(coll.ddata[k0]['data'][np.any(iokb, axis=axis)]),
            ]

    # --------------
    # Intermediate safety checks
    # --------------

    if np.any(np.isnan(data[iokb])):
        msg = (
            "Some NaNs in data not caught by iok!"
        )
        raise Exception(msg)

    if np.sum(iokb) == 0:
        msg = "There does not seem to be any usable data (no indok)"
        raise Exception(msg)

    # -----------------
    # validity vs nsigma
    # -----------------

    # Ok with and w/o binning if data provided as counts
    indv = np.zeros(data.shape, dtype=bool)
    indv[iokb] = np.sqrt(data[iokb]) > dvalid['nsigma']

    # update iokb

    # -----------------
    # validity vs faction
    # -----------------

    refi = coll.ddata[key_lamb]['ref'][0]
    axis = tuple([ii for ii, rr in enumerate(ref) if rr != refi])

    nlamb = coll.ddata[key_lamb]['data'].size

    axis = ref.index(refi)
    if dvalid.get('focus') is None:
        nlambi = np.sum(iokb, axis=axis, keepdims=True)
        frac = np.sum(indv, axis=axis, keepdims=True) / nlambi
        iout = frac < dvalid['fraction']

    else:
        shape = list(data.shape)
        shape[axis] = 1
        shape.append(len(dvalid['focus']))
        iout = np.zeros(tuple(shape), dtype=bool)
        frac = np.zeros(tuple(shape), dtype=float)
        nl = np.zeros(tuple(shape), dtype=int)
        sli = [slice(None) for ii in ref]
        for ii, ff in enumerate(dvalid['focus']):
            ilambok = (lamb >= ff[0]) & (lamb <= ff[1])
            nlambi = np.sum(ilambok)
            sli[axis] = ilambok
            frac[..., ii] = (
                np.sum(indv[tuple(sli)], axis=axis, keepdims=True) / nlambi
            )
            nl[..., ii] = nlambi

        if dvalid['focus_logic'] == 'min':
            frac = np.min(frac, axis=-1)
        elif dvalid['focus_logic'] == 'max':
            frac = np.max(frac, axis=-1)
        else:
            frac = np.sum(frac * nl, axis=-1) / np.sum(nl, axis=-1)

        iout = frac < dvalid['fraction']

    # reshape
    iout = np.repeat(iout, nlamb, axis=axis)

    # valid, excluded
    iokb = indv & iout
    iok[iokb] = -4

    # non-valid, included
    iokb = (~indv) & (iok == 0) & (~iout)
    iok[iokb] = -5

    # non-valid, excluded
    iokb = (~indv) & (iok == 0) & (iout)
    iok[iokb] = -6

    # -----------------
    # validity vs bspline mesh
    # -----------------

    if key_bs is not None:

        raise NotImplementedError()


    # ------------------------
    # more backup
    # ------------------------

    """
    # Derive indt and optionally dphi and indknots
    indbs, ldphi = False, False
    if focus is False:
        lambok = np.ones(tuple(np.r_[lamb.shape, 1]), dtype=bool)
        indall = ind[..., None]
    else:
        # TBC
        lambok = np.rollaxis(
            np.array([np.abs(lamb - ff[0]) < ff[1] for ff in focus]),
            0,
            lamb.ndim + 1,
        )
        indall = ind[..., None] & lambok[None, ...]
    nfocus = lambok.shape[-1]

    # -----------
    # backup

    if data2d is True:
        # Code ok with and without binning :-)

        # Get knots intervals that are ok
        fract = np.full((nspect, knots.size-1, nfocus), np.nan)
        for ii in range(knots.size - 1):
            iphi = (phi >= knots[ii]) & (phi < knots[ii + 1])
            fract[:, ii, :] = (
                np.sum(np.sum(indall & iphi[None, ..., None],
                              axis=1), axis=1)
                / np.sum(np.sum(iphi[..., None] & lambok,
                                axis=0), axis=0)
            )
        indknots = np.all(fract > valid_fraction, axis=2)

        # Deduce ldphi
        ldphi = [[] for ii in range(nspect)]
        for ii in range(nspect):
            for jj in range(indknots.shape[1]):
                if indknots[ii, jj]:
                    if jj == 0 or not indknots[ii, jj-1]:
                        ldphi[ii].append([knots[jj]])
                    if jj == indknots.shape[1] - 1:
                        ldphi[ii][-1].append(knots[jj+1])
                else:
                    if jj > 0 and indknots[ii, jj-1]:
                        ldphi[ii][-1].append(knots[jj])

        # Safety check
        assert all([
            all([len(dd) == 2 and dd[0] < dd[1] for dd in ldphi[ii]])
            for ii in range(nspect)
        ])

        # Deduce indbs that are ok
        nintpbs = nknotsperbs - 1
        indbs = np.zeros((nspect, nbs), dtype=bool)
        for ii in range(nbs):
            ibk = np.arange(max(0, ii-(nintpbs-1)), min(knots.size-1, ii+1))
            indbs[:, ii] = np.any(indknots[:, ibk], axis=1)

        assert np.all(
            (np.sum(indbs, axis=1) == 0) | (np.sum(indbs, axis=1) >= deg + 1)
        )

        # Deduce indt
        indt = np.any(indbs, axis=1)


    # ------------------------
    # more backup
    # ------------------------

    # Update indok with non-valid phi
    # non-valid = ok but out of dphi
    for ii in range(dinput['dprepare']['indok'].shape[0]):
        iphino = dinput['dprepare']['indok'][ii, ...] == 0
        for jj in range(len(dinput['valid']['ldphi'][ii])):
            iphino &= (
                (
                    dinput['dprepare']['phi']
                    < dinput['valid']['ldphi'][ii][jj][0]
                )
                | (
                    dinput['dprepare']['phi']
                    >= dinput['valid']['ldphi'][ii][jj][1]
                )
            )

        # valid, but excluded (out of dphi)
        iphi = (
            (dinput['dprepare']['indok'][ii, ...] == 0)
            & (dinput['valid']['ind'][ii, ...])
            & (iphino)
        )
        dinput['dprepare']['indok'][ii, iphi] = -5

        # non-valid, included (in dphi)
        iphi = (
            (dinput['dprepare']['indok'][ii, ...] == 0)
            & (~dinput['valid']['ind'][ii, ...])
            & (~iphino)
        )
        dinput['dprepare']['indok'][ii, iphi] = -6

        # non-valid, excluded (out of dphi)
        iphi = (
            (dinput['dprepare']['indok'][ii, ...] == 0)
            & (~dinput['valid']['ind'][ii, ...])
            & (iphino)
        )
        dinput['dprepare']['indok'][ii, iphi] = -7

    # indok_bool True if indok == 0 or -5 (because ...)
    dinput['dprepare']['indok_bool'] = (
        (dinput['dprepare']['indok'] == 0)
        | (dinput['dprepare']['indok'] == -6)
    )

    # add lambmin for bck
    dinput['lambmin_bck'] = np.min(dinput['dprepare']['lamb'])
    """

    # -----------------
    # store
    # -----------------

    # iok
    kiok = f"{key}_iok"
    dvalid['iok'] = kiok
    coll.add_data(
        key=kiok,
        data=iok,
        ref=ref,
        unit='',
        dim='bool',
    )

    # frac
    if data.ndim == 1:
        dvalid['frac'] = frac

    else:
        kfrac = f"{key}_frac"
        dvalid['frac'] = kfrac
        ref_lamb = coll.ddata[key_lamb]['ref'][0]
        ref_frac = tuple([rr for rr in ref if rr != ref_lamb])
        coll.add_data(
            key=kfrac,
            data=frac.squeeze(),
            ref=ref_frac,
            unit='',
            dim='bool',
        )

    return dvalid


#############################################
#############################################
#       validity - check dvalid
#############################################


def _check_dvalid_valid(
    dvalid=None,
    key_lamb=None,
):

    if not isinstance(dvalid, dict):
        msg = "Arg dvalid must be a dict\nProvided:\n{dvalid}"
        raise Exception(msg)

    # ---------------
    # positive
    # ---------------

    dvalid['positive'] = ds._generic_check._check_var(
        dvalid.get('positive'), "dvalid['positive']",
        types=bool,
        default=True,
    )

    # ---------------
    # update_domain
    # ---------------

    dvalid['update_domain'] = ds._generic_check._check_var(
        dvalid.get('update_domain'), "dvalid['update_domain']",
        types=bool,
        default=True,
    )

    # ---------------
    # nsigma
    # ---------------

    dvalid['nsigma'] = float(ds._generic_check._check_var(
        dvalid.get('nsigma'), "dvalid['nsigma']",
        types=(int, float),
        default=10,
        sign=">=0",
    ))

    # ---------------
    # fraction
    # ---------------

    dvalid['fraction'] = float(ds._generic_check._check_var(
        dvalid.get('fraction'), "dvalid['fraction']",
        types=(int, float),
        default=0.51,
        sign=">=0",
    ))

    # ---------------
    # focus
    # ---------------

    if dvalid.get('focus') is not None:

        dvalid['focus'] = _check_domain(
            domain=dvalid['focus'],
            key_lamb=key_lamb,
            key_bs_vect=None,
        )[key_lamb]['spec']

        # logic
        dvalid['focus_logic'] = ds._generic_check._check_var(
            dvalid.get('focus_logic'), "dvalid['focus_logic']",
            types=str,
            default='min',
            allowed=['min', 'max', 'sum'],
        )

    return dvalid


# -*- coding: utf-8 -*-
"""
Created on Tue Feb 20 14:44:51 2024

@author: dvezinet
"""


import datastock as ds


import numpy as np


# local
from . import _class01_valid as _valid


#############################################
#############################################
#       fit CHECK
#############################################


def _check(
    coll=None,
    # key
    key=None,
    # model
    key_model=None,
    # data and noise
    key_data=None,
    key_sigma=None,
    absolute_sigma=None,
    # wavelength and phi
    key_lamb=None,
    key_bs_vect=None,
    key_bs=None,
    # dict
    dparams=None,
    dvalid=None,
    # compute options
    chain=None,
):

    # ---------------------
    # basic check on inputs
    # ---------------------

    (
        key,
        key_model,
        key_data,
        key_sigma,
        key_lamb,
        key_bs_vect,
        key_bs,
        absolute_sigma,
        # derived
        ref,
        ref0,
        shape,
        shape0,
        axis_lamb,
        axis_bs,
    ) = _check_keys(**locals())

    # ---------------------
    # domain
    # ---------------------

    # prepare ddata
    ddata = {
        key_lamb: {
            'data': coll.ddata[key_lamb]['data'],
            'ref': coll.ddata[key_lamb]['ref'][0],
        }
    }
    if key_bs is not None:
        ddata.update({
            key_bs_vect: {
                'data': coll.ddata[key_bs_vect]['data'],
                'ref': coll.ddata[key_bs_vect]['ref'][0],
            },
        })

    # ---------------------
    # mask & domain
    # ---------------------

    dvalid = _valid.mask_domain(
        # resources
        coll=coll,
        key_data=key_data,
        key_lamb=key_lamb,
        key_bs_vect=key_bs_vect,
        # options
        dvalid=dvalid,
        ref=ref,
        ref0=ref0,
        shape0=shape0,
    )

    # ---------------------
    # validity
    # ---------------------

    dvalid = _valid.valid(
        coll=coll,
        key=key,
        key_data=key_data,
        key_lamb=key_lamb,
        key_bs=key_bs,
        dvalid=dvalid,
        ref=ref,
        ref0=ref0,
    )

    # ---------------------
    # dparams
    # ---------------------

    # if dparams is None:
        # dparams = _dparams.main()

    # # -------- BACKUP ------------
    # # Add dscales, dx0 and dbounds

    # dinput['dscales'] = fit12d_dscales(dscales=dscales, dinput=dinput)
    # dinput['dbounds'] = fit12d_dbounds(dbounds=dbounds, dinput=dinput)
    # dinput['dx0'] = fit12d_dx0(dx0=dx0, dinput=dinput)
    # dinput['dconstants'] = fit12d_dconstants(
        # dconstants=dconstants,
        # dinput=dinput,
    # )

    # ---------------------
    # store
    # ---------------------

    wsf = coll._which_fit
    dobj = {
        wsf: {
            key: {
                'key_model': key_model,
                'key_data': key_data,
                'key_sigma': key_sigma,
                'key_lamb': key_lamb,
                'key_bs': key_bs,
                'key_bs_vect': key_bs_vect,
                'key_sol': None,
                'key_cov': None,
                'absolute_sigma': absolute_sigma,
                'dparams': dparams,
                'dvalid': dvalid,
            },
        },
    }

    coll.update(dobj=dobj)

    return


#############################################
#############################################
#        check keys
#############################################


def _check_keys(
    coll=None,
    # keys
    key=None,
    key_model=None,
    key_data=None,
    key_sigma=None,
    key_lamb=None,
    key_bs_vect=None,
    key_bs=None,
    absolute_sigma=None,
    # unused
    **kwdargs,
):

    # -------------
    # key
    # -------------

    wsf = coll._which_fit
    key = ds._generic_check._obj_key(
        d0=coll.dobj.get(wsf, {}),
        short='sf',
        key=key,
        ndigits=2,
    )

    # -------------
    # key_model
    # -------------

    wsm = coll._which_model
    lok = list(coll.dobj.get(wsm, {}).keys())
    key_model = ds._generic_check._check_var(
        key_model, 'key_model',
        types=str,
        allowed=lok,
    )

    # -------------
    # key_data
    # -------------

    # key_data
    lok = list(coll.ddata.keys())
    key_data = ds._generic_check._check_var(
        key_data, 'key_data',
        types=str,
        allowed=lok,
    )

    # derive refs
    ref = coll.ddata[key_data]['ref']
    shape = coll.ddata[key_data]['shape']

    # -------------
    # key_lamb
    # -------------

    # key_lamb
    lok = [
        k0 for k0, v0 in coll.ddata.items()
        if v0['monot'] == (True,)
        and v0['ref'][0] in ref
    ]
    key_lamb = ds._generic_check._check_var(
        key_lamb, 'key_lamb',
        types=str,
        allowed=lok,
        extra_msg="Should be a 1d vector, strictly monotonous",
    )

    # axis_lamb
    ref_lamb = coll.ddata[key_lamb]['ref'][0]
    axis_lamb = ref.index(ref_lamb)

    # -------------
    # key_sigma
    # -------------

    if key_sigma is None:
        key_sigma = 'poisson'

    # key_sigma
    if isinstance(key_sigma, str):

        # percentage
        c0 = (
            key_sigma[-1] == '%'
            and all([ss.isnumeric() or ss == '.' for ss in key_sigma[:-1]])
        )
        if c0:
            asig_def = True

        else:

            lok_data = [
                k0 for k0, v0 in coll.ddata.items()
                if v0['units'] == coll.ddata[key_data]['units']
                and (
                    v0['ref'] == ref
                    or v0['ref'] == ref_lamb
                )
                and np.all(v0['data'] > 0)
            ]

            lok = list(coll.ddata.keys())
            key_sigma = ds._generic_check._check_var(
                key_sigma, 'key_sigma',
                types=str,
                allowed=lok_data + ['poisson'],
                default='poisson',
                extra_msg=_err_key_sigma(key_sigma, lok=lok, returnas=True),
            )

            asig_def = True

    else:
        key_sigma = float(ds._generic_check._check_var(
            key_sigma, 'key_sigma',
            types=(int, float),
            sign='>0',
            extra_msg=_err_key_sigma(key_sigma, returnas=True),
        ))
        asig_def = True

    # -------------
    # absolute_sigma
    # -------------

    absolute_sigma = ds._generic_check._check_var(
        absolute_sigma, 'absolute_sigma',
        types=bool,
        default=asig_def,
        extra_msg=(
            "Determines whether the error 'key_sigma' should be "
            "understood as a relative or absolute error bar"
        ),
    )

    # -------------
    # key_bs
    # -------------

    c0 = (
        len(ref) >= 2
        and key_bs_vect is not None
    )
    if c0:

        # key_bs_vect
        lok = [
            k0 for k0, v0 in coll.ddata.items()
            if v0['monot'] == (True,)
            and v0['ref'][0] in ref
            and v0['ref'][0] != coll.ddata[key_lamb]['ref'][0]
        ]
        key_bs_vect = ds._generic_check._check_var(
            key_bs_vect, 'key_bs_vect',
            types=str,
            allowed=lok,
        )

        axis_bs = ref.index(coll.ddata[key_bs_vect]['ref'][0])

        # units, dim
        units = coll.ddata[key_bs_vect]['units']
        quant = coll.ddata[key_bs_vect]['quant']
        dim = coll.ddata[key_bs_vect]['dim']

        # key_bs
        wbs = coll._which_bsplines
        lok = [
            k0 for k0, v0 in coll.dobj.get(wbs, {}).items()
            if len(v0['shape']) == 1
            and (
                coll.ddata[v0['apex'][0]]['units'] == units
                or coll.ddata[v0['apex'][0]]['quant'] == quant
                or coll.ddata[v0['apex'][0]]['dim'] == dim
            )
        ]
        key_bs = ds._generic_check._check_var(
            key_bs, 'key_bs',
            types=str,
            allowed=lok,
        )

        # ref0
        ref0 = tuple([
            rr for ii, rr in enumerate(ref)
            if ii in [axis_lamb, axis_bs]
        ])

        # shape0
        shape0 = tuple([
            ss for ii, ss in enumerate(shape)
            if ii in [axis_lamb, axis_bs]
        ])

    else:
        key_bs_vect = None
        key_bs = None
        axis_bs = None
        shape0 = (shape[axis_lamb],)
        ref0 = (ref_lamb,)

    return (
        key,
        key_model,
        key_data,
        key_sigma,
        key_lamb,
        key_bs_vect,
        key_bs,
        absolute_sigma,
        # derived
        ref,
        ref0,
        shape,
        shape0,
        axis_lamb,
        axis_bs,
    )


def _err_key_sigma(key_sigma=None, lok=None, returnas=False):

    msg = (
        "Arg 'key_sigma' must either be:\n"
        "\t- str: a key to a array with:\n"
            "\t\t- same units as key_data\n"
            "\t\t- same ref as key_data or key_lamb\n"
            "\t\t- only strictly positive values\n"
    )
    if lok is not None:
        msg += f"\t\tAvailable:\n\t{lok}\n"

    msg += (
        "\t- 'poisson': poisson statictics (sqrt(data))\n"
        "\t- float or int : constant unique sigma for all data points\n"
        "\t- str: a float with '%' (e.g.: '5.0%'), constant percentage error\n"
        f"\nProvided:\n\t{key_sigma}"
    )

    if returnas is True:
        return msg
    else:
        raise Exception(msg)


# ###########################################
# ###########################################
#        check dinitial
# ###########################################


# def _check_dscales(
    # coll=None,
    # key_model=None,
    # key_data=None,
    # key_lamb=None,
    # dscales=None,
# ):


# -*- coding: utf-8 -*-
"""
Created on Tue Feb 20 19:01:39 2024

@author: dvezinet
"""


import copy


import itertools as itt
import numpy as np
import datastock as ds


#############################################
#############################################
#       main
#############################################

"""
def main(
    coll=None,
    dparams=None,
    key_model=None,
    key_data=None,
    key_lamb=None,
    key_bs=None,
):

    # ------------
    # check inputs
    # ------------

    dparams = _check(dparams)

    # ------------
    # compute
    # ------------

    # ------------------
    # list all variables
    lvar = coll.get_spectral_model_variables(
        key=key_model,
        all_free_tied='all',
        concatenate=True,
    )

    # ----------------------
    # prepare data if needed

    c0 = any([dparams.get(k0) is None for k0 in lvar])
    if c0:

        lamb = coll.ddata[key_lamb]['data']
        data = coll.ddata[key_data]['data']

        # TBF
        Dlamb = np.diff(dinput['dprepare']['domain']['lamb']['minmax'])
        lambm = dinput['dprepare']['domain']['lamb']['minmax'][0]
        if not (np.isfinite(Dlamb)[0] and Dlamb > 0):
            msg = (
                "lamb min, max seems to be non-finite or non-positive!\n"
                + "\t- dinput['dprepare']['domain']['lamb']['minmax'] = {}".format(
                    dinput['dprepare']['domain']['lamb']['minmax']
                )
                + "\n  => Please provide domain['lamb']"
            )
            raise Exception(msg)
        if lambm == 0:
            lambm = Dlamb / 100.

    # ----------------------
    # compute if needed

    # check all variables
    for k0 in lvar:

        if dparams.get(k0) is None:

            key, var = k0.split('_')
            typ = coll.dobj[key_model]['dmodel'][key]['type']

            scale = _get_scale(
                key_model=key_model,
                key=k0,
                typ=typ,
                var=var,
                # data
                lamb=lamb,
                data=data,
            )

            dparams[k0] = scale

    return


#############################################
#############################################
#       check
#############################################


def _check(
    coll=None,
    dparams=None,
    key_model=None,
    key_data=None,
    key_lamb=None,
    key_bs=None,
):

    # --------------------
    # get model variables
    # --------------------

    lvar = coll.get_spectral_model_variables(
        key=key_model,
        all_free_tied='all',
        concatenate=True,
    )

    # --------------------
    # Trivial: False = 1
    # --------------------

    if dparams is False:
        dparams = {k0: 1. for k0 in lvar}
        return dparams

    # --------------------
    # Trivial: scalar => uniform
    # --------------------

    if np.isscalar(dparams):
        dparams = {k0: dparams for k0 in lvar}
        return dparams

    # --------------------
    # Non-trivial
    # --------------------

    # None
    if dparams is None:
        dparams = {}

    c0 = (
        isinstance(dparams, dict)
        and all([
            k0 in lvar
            and ((np.isscalar(v0) and np.isfinite(v0)) or v0 is None)
            for k0, v0 in dparams.items()
        ])
    )
    if not c0:
        msg = (
            "Arg dparams must be a dict with keys = variables of model"
            " and values = floats\n"
            f"\t- model: {key_model}\n"
            f"\t- variables: {lvar}\n"
            f"Provided:\n{dparams}\n"
        )
        raise Exception(msg)

    return dparams


#############################################
#############################################
#       get scale from data
#############################################


def _get_scale(
    key_model=None,
    key=None,
    typ=None,
    var=None,
):

    # ------------
    # bck
    # ------------

    if typ == 'poly':

        if var == 'c0':
            pass

        if var == 'c1':
            pass

        else:
            err = True

    # ------------
    # exp
    # ------------

    elif typ == 'exp':

        indbck = (data > np.nanmean(data, axis=1)[:, None]) | (~indok)
        bcky = np.array(np.ma.masked_where(indbck, data).mean(axis=1))
        bckstd = np.array(np.ma.masked_where(indbck, data).std(axis=1))

        iok = (bcky > 0) & (bckstd > 0)
        if (bck_rate is None or nbck_amp is None) and not np.any(iok):
            bcky = 0.1*np.array(np.ma.masked_where(~indbck, data).mean(axis=1))
            bckstd = 0.1*bcky
        elif not np.all(iok):
            bcky[~iok] = np.mean(bcky[iok])
            bckstd[~iok] = np.mean(bckstd[iok])

        # bck_rate
        if bck_rate is None:
            bck_rate = (
                np.log((bcky + bckstd)/bcky) / (lamb.max()-lamb.min())
            )
        if bck_amp is None:
            # Assuming bck = A*exp(rate*(lamb-lamb.min()))
            bck_amp = bcky

        dparams = _fit12d_filldef_dparamsx0_float(
            din=dparams, din_name='dparams', key='bck_amp',
            vref=bck_amp, nspect=nspect,
        )

        if var == 'amp':
            pass

        elif var == 'rate':
            pass

        else:
            err = True

    # ------------
    # gauss
    # ------------

    elif typ == 'gauss':

        if var == 'amp':
            pass

        elif var == 'w2':
            pass

        elif var == 'x0':
            pass

        else:
            err = True

    # ------------
    # lorentz
    # ------------

    elif typ == 'lorentz':

        if var == 'amp':
            pass

        elif var == 'gamma':
            pass

        elif var == 'x0':
            pass

        else:
            err = True

    # ------------
    # pvoigt
    # ------------

    elif typ == 'pvoigt':

        if var == 'amp':
            pass

        elif var == 'gamma':
            pass

        elif var == 'x0':
            pass

        elif var == 'w2':
            pass

        elif var == 't':
            pass

        else:
            err = True

    # ------------
    # voigt
    # ------------

    # ------------
    # error
    # ------------

    if err is True:
        msg = (
            f"Spectral model '{key_model}', issue with variable '{key}':\n"
            f"unknown variable '{var}' for type '{typ}'"
        )
        raise Exception(msg)

    return val


#############################################
#############################################
#       Backup
#############################################


def _fit12d_filldef_dscalesx0_dict(
    din=None, din_name=None,
    key=None, vref=None,
    nspect=None, dinput=None,
):

    # Check vref
    if vref is not None:
        if type(vref) not in _LTYPES and len(vref) not in [1, nspect]:
            msg = (
                "Non-conform vref for "
                + "{}['{}']\n".format(din_name, key)
                + "\t- expected: float or array (size {})\n".format(nspect)
                + "\t- provided: {}".format(vref)
            )
            raise Exception(msg)
        if type(vref) in _LTYPES:
            vref = np.full((nspect,), vref)
        elif len(vref) == 1:
            vref = np.full((nspect,), vref[0])

    # check din[key]
    if din.get(key) is None:
        assert vref is not None
        din[key] = {k0: vref for k0 in dinput[key]['keys']}

    elif not isinstance(din[key], dict):
        assert type(din[key]) in _LTYPES + [np.ndarray]
        if hasattr(din[key], '__len__') and len(din[key]) == 1:
            din[key] = din[key][0]
        if type(din[key]) in _LTYPES:
            din[key] = {
                k0: np.full((nspect,), din[key])
                for k0 in dinput[key]['keys']
            }
        elif din[key].shape == (nspect,):
            din[key] = {k0: din[key] for k0 in dinput[key]['keys']}
        else:
            msg = (
                "{}['{}'] not conform!".format(dd_name, key)
            )
            raise Exception(msg)

    else:
        for k0 in dinput[key]['keys']:
            if din[key].get(k0) is None:
                din[key][k0] = vref
            elif type(din[key][k0]) in _LTYPES:
                din[key][k0] = np.full((nspect,), din[key][k0])
            elif len(din[key][k0]) == 1:
                din[key][k0] = np.full((nspect,), din[key][k0][0])
            elif din[key][k0].shape != (nspect,):
                msg = (
                    "Non-conform value for "
                    + "{}['{}']['{}']\n".format(din_name, key, k0)
                    + "\t- expected: float or array (size {})\n".format(nspect)
                    + "\t- provided: {}".format(din[key][k0])
                )
                raise Exception(msg)
    return din


def _fit12d_filldef_dscalesx0_float(
    din=None, din_name=None,
    key=None, vref=None,
    nspect=None,
):
    if din.get(key) is None:
        if type(vref) in _LTYPES:
            din[key] = np.full((nspect,), vref)
        elif np.array(vref).shape == (1,):
            din[key] = np.full((nspect,), vref[0])
        elif np.array(vref).shape == (nspect,):
            din[key] = np.array(vref)
        else:
            msg = (
                "Non-conform vref for {}['{}']\n".format(din_name, key)
                + "\t- expected: float or array (size {})\n".format(nspect)
                + "\t- provided: {}".format(vref)
            )
            raise Exception(msg)
    else:
        if type(din[key]) in _LTYPES:
            din[key] = np.full((nspect,), din[key])
        elif din[key].shape == (1,):
            din[key] = np.full((nspect,), din[key][0])
        elif din[key].shape != (nspect,):
            msg = (
                "Non-conform vref for {}['{}']\n".format(din_name, key)
                + "\t- expected: float or array (size {})\n".format(nspect)
                + "\t- provided: {}".format(din[key])
            )
            raise Exception(msg)
    return din


def _check_finit_dict(dd=None, dd_name=None, indtok=None, indbs=None):
    dfail = {}
    for k0, v0 in dd.items():
        if k0 in ['amp', 'width', 'shift']:
            for k1, v1 in v0.items():
                if np.any(~np.isfinite(v1[indtok, ...])):
                    dfail[f"'{k0}'['{k1}']"] = v1
        elif k0 == 'bs':
            if np.any(~np.isfinite(v0[indbs])):
                dfail[f"'{k0}'"] = v0
        else:
            if np.any(~np.isfinite(v0[indtok, ...])):
                dfail[f"'{k0}'"] = v0

    if len(dfail) > 0:
        lstr = [f"\t- {k0}: {v0}" for k0, v0 in dfail.items()]
        msg = (
            f"The following {dd_name} values are non-finite:\n"
            + "\n".join(lstr)
        )
        raise Exception(msg)



# Double-check 1d vs 2d: TBF / TBC
def fit12d_dscales(dscales=None, dinput=None):

    # --------------
    # Input checks
    dscales = _fit12d_checkformat_dscalesx0(
        din=dscales, dinput=dinput, name='dscales',
    )

    data = dinput['dprepare']['data']
    lamb = dinput['dprepare']['lamb']
    nspect = data.shape[0]

    # --------------
    # 2d spectrum = 1d spectrum + vert. profile
    is2d = data.ndim == 3
    if is2d is True:
        data = dinput['dprepare']['datalamb1d']
        datavert = dinput['dprepare']['dataphi1d']
        lamb = dinput['dprepare']['lamb1d']
        phi = dinput['dprepare']['phi1d']
        indok = np.any(dinput['dprepare']['indok_bool'], axis=1)

        # bsplines modulation of bck and amp, if relevant
        # fit bsplines on datavert (vertical profile)
        # to modulate scales (bck and amp)

        if dinput['symmetry'] is True:
            phitemp = np.abs(phi[None, :] - dinput['symmetry_axis'][:, None])
        else:
            phitemp = np.tile(phi, (nspect, 1))

        # Loop on time and bsplines
        dscales['bs'] = np.full((nspect, dinput['nbs']), np.nan)
        for ii in dinput['valid']['indt'].nonzero()[0]:
            for jj, jbs in enumerate(range(dinput['nbs'])):
                if dinput['valid']['indbs'][ii, jj]:
                    kn0 = dinput['knots_mult'][jj]
                    kn1 = dinput['knots_mult'][jj + dinput['nknotsperbs'] - 1]
                    indj = (
                        (~np.isnan(datavert[ii, :]))
                        & (kn0 <= phitemp[ii, :])
                        & (phitemp[ii, :] <= kn1)
                    )
                    if not np.any(indj):
                        msg = "Unconsistent indbs!"
                        raise Exception(msg)
                    dscales['bs'][ii, jj] = np.mean(datavert[ii, indj])

        # Normalize to avoid double-amplification when amp*bs
        corr = np.nanmax(dscales['bs'][dinput['valid']['indt'], :], axis=1)
        dscales['bs'][dinput['valid']['indt'], :] /= corr[:, None]
    else:
        indok = dinput['dprepare']['indok_bool']

    # --------------
    # Default values for filling missing fields
    Dlamb = np.diff(dinput['dprepare']['domain']['lamb']['minmax'])
    lambm = dinput['dprepare']['domain']['lamb']['minmax'][0]
    if not (np.isfinite(Dlamb)[0] and Dlamb > 0):
        msg = (
            "lamb min, max seems to be non-finite or non-positive!\n"
            + "\t- dinput['dprepare']['domain']['lamb']['minmax'] = {}".format(
                dinput['dprepare']['domain']['lamb']['minmax']
            )
            + "\n  => Please provide domain['lamb']"
        )
        raise Exception(msg)
    if lambm == 0:
        lambm = Dlamb / 100.

    # bck_amp
    bck_amp = dscales.get('bck_amp')
    bck_rate = dscales.get('bck_rate')
    if bck_amp is None or bck_rate is None:
        indbck = (data > np.nanmean(data, axis=1)[:, None]) | (~indok)
        bcky = np.array(np.ma.masked_where(indbck, data).mean(axis=1))
        bckstd = np.array(np.ma.masked_where(indbck, data).std(axis=1))

        iok = (bcky > 0) & (bckstd > 0)
        if (bck_rate is None or nbck_amp is None) and not np.any(iok):
            bcky = 0.1*np.array(np.ma.masked_where(~indbck, data).mean(axis=1))
            bckstd = 0.1*bcky
        elif not np.all(iok):
            bcky[~iok] = np.mean(bcky[iok])
            bckstd[~iok] = np.mean(bckstd[iok])

        # bck_rate
        if bck_rate is None:
            bck_rate = (
                np.log((bcky + bckstd)/bcky) / (lamb.max()-lamb.min())
            )
        if bck_amp is None:
            # Assuming bck = A*exp(rate*(lamb-lamb.min()))
            bck_amp = bcky

    dscales = _fit12d_filldef_dscalesx0_float(
        din=dscales, din_name='dscales', key='bck_amp',
        vref=bck_amp, nspect=nspect,
    )
    dscales = _fit12d_filldef_dscalesx0_float(
        din=dscales, din_name='dscales', key='bck_rate',
        vref=bck_rate, nspect=nspect,
    )

    # amp
    dscales['amp'] = dscales.get('amp', dict.fromkeys(dinput['amp']['keys']))
    for ii, ij in enumerate(dinput['dind']['amp_x0']):
        key = dinput['amp']['keys'][ii]
        if dscales['amp'].get(key) is None:
            # convoluate and estimate geometric mean
            conv = np.exp(
                    -(lamb - dinput['lines'][ij])**2 / (2*(Dlamb / 25.)**2)
                )[None, :]
            dscales['amp'][key] = np.nanmax(data*conv, axis=1)
        else:
            if type(dscales['amp'][key]) in _LTYPES:
                dscales['amp'][key] = np.full((nspect,), dscales['amp'][key])
            else:
                assert dscales['amp'][key].shape == (nspect,)

    # width
    if dinput.get('same_spectrum') is True:
        lambm2 = (
            lambm
            + dinput['same_spectrum_dlamb']
            * np.arange(0, dinput['same_spectrum_nspect'])
        )
        nw0 = iwx.size / dinput['same_spectrum_nspect']
        lambmw = np.repeat(lambm2, nw0)
        widthref = (Dlamb/(25*lambmw))**2
    else:
        widthref = (Dlamb/(25*lambm))**2

    dscales = _fit12d_filldef_dscalesx0_dict(
        din=dscales, din_name='dscales', key='width', vref=widthref,
        nspect=nspect, dinput=dinput,
    )

    # shift
    shiftref = Dlamb/(25*lambm)
    dscales = _fit12d_filldef_dscalesx0_dict(
        din=dscales, din_name='dscales', key='shift', vref=shiftref,
        nspect=nspect, dinput=dinput,
    )

    # Double
    if dinput['double'] is not False:
        dratio = 1.
        dshift = float(Dlamb/(25*lambm))
        if dinput['double'] is True:
            pass
        else:
            if dinput['double'].get('dratio') is not None:
                dratio = dinput['double']['dratio']
            if dinput['double'].get('dshift') is not None:
                dshift = dinput['double']['dshift']
        din = {'dratio': dratio, 'dshift': dshift}
        for k0 in din.keys():
            dscales = _fit12d_filldef_dscalesx0_float(
                din=dscales, din_name='dscales', key=k0,
                vref=din[k0], nspect=nspect,
            )
    elif 'dratio' in dscales.keys():
        del dscales['dratio'], dscales['dshift']

    # check
    _check_finit_dict(
        dd=dscales,
        dd_name='dscales',
        indtok=dinput['valid']['indt'],
        indbs=dinput['valid']['indbs'],
    )
    return dscales
"""


# -*- coding: utf-8 -*-


import os
import itertools as itt


import numpy as np
import scipy.constants as scpct
import matplotlib.pyplot as plt


# ###################################################
# ###################################################
#               DEFAULTS
# ###################################################


# PATH
_PATH_HERE = os.path.dirname(__file__)
_PATH_INPUT = os.path.join(_PATH_HERE, 'input')


# PFE
_MASK_1D = os.path.join(_PATH_INPUT, 'mask1d.npy')


# LAMB
_NLANB = 300
_LAMB = lamb = np.linspace(3.9, 4, _NLANB)*1e-10
_LAMB0 = _LAMB[0] + (_LAMB[-1] - _LAMB[0]) * np.r_[0.25, 0.55, 0.75]


# ###################################################
# ###################################################
#               data
# ###################################################


def add_data(coll=None, lamb=_LAMB, lamb0=_LAMB0):

    # ------------------
    # reference vectors
    # ------------------

    # lamb
    coll.add_ref('nlamb', size=lamb.size)
    coll.add_data(
        'lamb',
        data=lamb,
        ref='nlamb',
        units='m',
        dim='dist',
        quant='wavelength',
    )

    # derive useful quantities
    Dlamb = lamb[-1] - lamb[0]
    lambm = 0.5*(lamb[-1] + lamb[0])

    # phi
    nphi = 100
    coll.add_ref('nphi', size=nphi)
    phi = np.linspace(-0.1, 0.1, nphi)
    coll.add_data(
        'phi',
        data=phi,
        ref='nphi',
        units='rad',
        dim='angle',
        quant='phi',
    )

    # time
    nt = 50
    coll.add_ref('nt', size=nt)
    t = np.linspace(0, 10, nt)
    coll.add_data(
        't',
        data=t,
        ref='nt',
        units='s',
        dim='time',
        quant='t',
    )

    # ------------------
    # model-specific data
    # ------------------

    # --------
    # poly

    lamb_rel = (lamb - lambm) / Dlamb

    a0 = 5000
    a1 = 1500
    a2 = -3000

    coll.add_data(
        key='data_poly',
        data=np.random.poisson(a0 + a1*lamb_rel + a2*lamb_rel**2),
        ref='nlamb',
        units='counts',
    )

    # --------
    # exp_lamb

    Te = 1e3
    rate = scpct.h * scpct.c / (Te * scpct.e)
    amp = 500 * lambm * np.exp(rate/lambm)

    coll.add_data(
        key='data_exp',
        data=np.random.poisson(amp * np.exp(-rate/lamb) / lamb),
        ref='nlamb',
        units='counts',
    )

    # --------
    # gauss

    sigma = Dlamb/20.
    vccos = Dlamb/10. / lamb0[0]
    amp = 100 * np.exp((lambm - lamb0[0]*(1+vccos))**2/(2*sigma**2))

    coll.add_data(
        key='data_gauss',
        data=np.random.poisson(
            amp * np.exp(-(lamb - lamb0[0]*(1+vccos))**2/(2*sigma**2))
        ),
        ref='nlamb',
        units='counts',
    )

    # --------
    # double gauss

    coll.add_data(
        key='data_gauss2',
        data=np.random.poisson(
            amp * np.exp(-(lamb - lamb0[0]*(1+vccos))**2/(2*sigma**2))
            + 0.5*amp * np.exp(-(lamb - lamb0[1]*(1+vccos))**2/(2*sigma**2))
        ),
        ref='nlamb',
        units='counts',
    )

    # --------
    # lorentz

    gam = Dlamb/20.
    vccos = Dlamb/10. / lamb0[0]
    amp = 100 * (1 + ((lambm - lamb0[0]*(1 + vccos)) / gam)**2)

    coll.add_data(
        key='data_lorentz',
        data=np.random.poisson(
            amp / (1 + ((lamb - lamb0[0]*(1 + vccos)) / gam)**2)
        ),
        ref='nlamb',
        units='counts',
    )

    # --------
    # double lorentz

    coll.add_data(
        key='data_lorentz2',
        data=np.random.poisson(
            amp / (1 + ((lamb - lamb0[0]*(1 + vccos)) / gam)**2)
            + 0.5* amp / (1 + ((lamb - lamb0[1]*(1 + vccos)) / gam)**2)
        ),
        ref='nlamb',
        units='counts',
    )

    # --------
    # pvoigt

    sigma = Dlamb/20.
    gam = Dlamb/20.
    vccos = Dlamb/10. / lamb0[1]
    eta = 0.2
    amp = 100 / (
        eta / (1 + ((lambm - lamb0[0]*(1 + vccos)) / gam)**2)
        + (1-eta) * np.exp(-(lambm - lamb0[0]*(1 + vccos))**2/(2*sigma**2))
    )

    pvoigt = amp * (
        eta / (1 + ((lamb - lamb0[0]*(1 + vccos)) / gam)**2)
        + (1-eta) * np.exp(
            -(lamb - lamb0[0]*(1 + vccos))**2
            / (2*sigma**2)
        )
    )

    coll.add_data(
        key='data_pvoigt',
        data=np.random.poisson(pvoigt),
        ref='nlamb',
        units='counts',
    )

    # --------
    # double pvoigt

    pvoigt2 = amp * (
        eta / (1 + ((lamb - lamb0[1]*(1 + vccos)) / gam)**2)
        + (1-eta) * np.exp(
            -(lamb - lamb0[1]*(1 + vccos))**2
            / (2*sigma**2)
        )
    )

    coll.add_data(
        key='data_pvoigt2',
        data=np.random.poisson(pvoigt + 0.5 * pvoigt2),
        ref='nlamb',
        units='counts',
    )

    # --------
    # pulse_exp

    t0 = lamb0[0]
    tup = Dlamb / 30
    tdown = Dlamb / 5
    amp = 100 / (np.exp(-(lambm-t0)/tdown) - np.exp(-(lambm-t0)/tup))

    coll.add_data(
        key='data_pulse_exp',
        data=np.random.poisson(
            amp * (lamb >= t0) * (
                np.exp(-(lamb-t0)/tdown) - np.exp(-(lamb-t0)/tup)
            )
        ),
        ref='nlamb',
        units='counts',
    )

    # --------
    # double pulse_exp

    deltat = 0.04e-10

    coll.add_data(
        key='data_pulse_exp2',
        data=np.random.poisson(
            amp * (lamb >= t0) * (
                np.exp(-(lamb-t0)/tdown) - np.exp(-(lamb-t0)/tup)
            )
            + 0.5*amp * (lamb >= (t0 + deltat)) * (
                np.exp(-(lamb-(t0 + deltat))/tdown) - np.exp(-(lamb-(t0 + deltat))/tup)
            )
        ),
        ref='nlamb',
        units='counts',
    )

    # --------
    # pulse_gauss

    t0 = lamb0[0]
    tup = Dlamb / 30
    tdown = Dlamb / 5
    indup = (lamb < t0)
    inddown = (lamb >= t0)
    amp = 100

    coll.add_data(
        key='data_pulse_gauss',
        data=np.random.poisson(
            amp * (
                indup * np.exp(-(lamb - t0)**2/tup**2)
                + inddown * np.exp(-(lamb - t0)**2/tdown**2)
            )
        ),
        ref='nlamb',
        units='counts',
    )

    # --------
    # double pulse_gauss

    coll.add_data(
        key='data_pulse_gauss2',
        data=np.random.poisson(
            amp * (
                indup * np.exp(-(lamb - t0)**2/tup**2)
                + inddown * np.exp(-(lamb - t0)**2/tdown**2)
            )
            + 0.5 * amp * (
                (lamb < (t0+deltat)) * np.exp(-(lamb - (t0+deltat))**2/tup**2)
                + (lamb >= (t0+deltat)) * np.exp(-(lamb - (t0+deltat))**2/tdown**2)
            )
        ),
        ref='nlamb',
        units='counts',
    )

    # --------
    # lognorm

    t0 = lamb0[0]
    sigma = 0.5
    mu = 0.5 * (np.log((Dlamb / 30)**2/(np.exp(sigma**2) - 1)) - sigma**2)

    amp = (
        100 * (lambm - t0)
        * np.exp((np.log(lambm - t0) - mu)**2 / (2.*sigma**2))
    )
    data = np.zeros((lamb.size,))
    iok = (lamb >= t0)
    data[iok] = (amp / (lamb[iok] - t0)) * np.exp(
        -(np.log(lamb[iok] - t0) - mu)**2 / (2.*sigma**2)
    )

    coll.add_data(
        key='data_lognorm',
        data=np.random.poisson(data),
        ref='nlamb',
        units='counts',
    )

    # --------
    # double lognorm

    iok = (lamb >= (t0 + deltat))
    data2 = np.zeros((lamb.size,))
    data2[iok] = (amp / (lamb[iok] - (t0 + deltat))) * np.exp(
        -(np.log(lamb[iok] - (t0 + deltat)) - mu)**2 / (2.*sigma**2)
    )

    coll.add_data(
        key='data_lognorm2',
        data=np.random.poisson(data + 0.5 * data2),
        ref='nlamb',
        units='counts',
    )

    # ------------------
    # data 1d
    # ------------------

    # lamb0
    amp0 = 700
    width = Dlamb * np.r_[0.015, 0.035, 0.025]
    amp = amp0 * np.r_[1, 0.5, 2]

    amp1 = np.max(amp) * 0.10
    amp2 = np.max(amp) * 0.02
    dlamb = -(lamb[0] - lamb[-1]) / np.log(amp1/amp2)
    A = amp1 / np.exp(-lamb[0]/dlamb)

    # data 1d
    data = np.random.poisson(
        A * np.exp(-lamb / dlamb)
        + np.sum(
            [
                amp[ii] * np.exp(-(lamb-lamb0[ii])**2 / (2*width[ii]**2))
                for ii in range(len(lamb0))
            ],
            axis=0,
        ),
        size=lamb.size,
    ).astype(float)     # + amp0 * 0.10 * np.random.random((nlamb,))

    # store
    coll.add_data(
        'data1d',
        data=data,
        ref='nlamb',
        units='ph',
    )

    # ------------------
    # data 2d (1d vs t)
    # ------------------

    ampt = np.exp(-(t-np.mean(t))**2 / (0.3*(t[-1] - t[0]))**2)
    At = A * ampt[:, None]
    lambt = lamb[None, :]
    dv = np.r_[0.1, -0.05, 0.08]

    # data 1d
    data = np.random.poisson(
        At * np.exp(-lambt / dlamb)
        + ampt[:, None] * np.sum(
            [
                amp[ii]
                * np.exp(
                    - (lambt-lamb0[ii] - dv[ii] * Dlamb * ampt[:, None])**2
                    / (2*ampt[:, None] * width[ii]**2)
                )
                for ii in range(len(lamb0))
            ],
            axis=0,
        ),
        size=(nt, lamb.size),
    ).astype(float)     # + amp0 * 0.10 * np.random.random((nlamb,))

    # store
    coll.add_data(
        'data2d',
        data=data,
        ref=('nt', 'nlamb'),
        units='ph',
    )

    # ------------------
    # data 2d (1d + bs)
    # ------------------

    # ------------------
    # data 3d (1d + bs + t)
    # ------------------

    return


# ####################################################
# ####################################################
#            Add model-specific data (xfree)
# ####################################################


def add_xfree(coll=None):

    # --------------
    # prepare xfree

    t = coll.ddata['t']['data']
    dxfree = _get_dxfree(t, lamb)

    # ---------------
    # add model data

    for ii, kmodel in enumerate(dxfree.keys()):

        # ref_nx
        ref_nx = f'nx_{kmodel}'

        # xfree
        xfree = dxfree[kmodel]
        if xfree.ndim == 2:
            ref = (coll.ddata['t']['ref'][0], ref_nx)
        else:
            ref = (ref_nx,)

        # add_data
        kdata = f"xfree_{kmodel}"
        coll.add_data(
            key=kdata,
            data=xfree,
            ref=ref,
        )

    return


def _get_dxfree(t=None, lamb=None):

    lambD = lamb[-1] - lamb[0]
    tm = np.mean(t)

    rate = np.log(2*lamb[0]/lamb[-1]) / (1/lamb[-1] - 1/lamb[0])

    tmax = 3.960e-10
    fmax = 1
    delta = lambD / 30

    sigma = 0.5
    mu = 0.5 * (np.log(delta**2/(np.exp(sigma**2) - 1)) - sigma**2)
    # mu = -28

    t0 = (tmax - np.exp(mu - sigma**2))
    tau = (t0 - lamb[0]) / lambD
    amp = fmax / np.exp(0.5*sigma**2 - mu)

    # sm00
    # 'bck0_a0', 'bck0_a1', 'bck_a2',
    # 'l00_amp', 'l00_vccos', 'l00_sigma',
    # 'l08_amp', 'l08_vccos', 'l08_gam',
    # 'l12_amp', 'l12_vccos', 'l12_sigma', 'l12_gam',

    # sm01
    # 'bck0_amp', 'bck0_rate',
    # 'l00_amp', 'l00_vccos', 'l00_sigma',
    # 'l01_vccos', 'l01_gam',
    # 'l02_amp', 'l02_vccos',

    # sm02
    # 'bck0_amp', 'bck0_rate',
    # 'l00_amp', 'l00_vccos', 'l00_sigma',
    # 'l01_vccos',
    # 'l02_amp', 'l02_vccos', 'l02_sigma', 'l02_gam'

    # sm03
    # 'bck0_a0', 'bck0_a1', 'bck_a2',
    # 'l00_amp', 'l00_tau', 'l00_t_up', 'l00_t_down',
    # 'l01_amp', 'l01_tau', 'l01_t_up', 'l01_t_down',
    # 'l02_amp', 'l02_tau', 'l02_mu', 'l02_sigma'

    dxfree = {
        # testing elementary models
        # 'smpoly': np.r_[0.1 - lamb[0] * 0.1/lambD, 0.1 / lambD],
        'smpoly': np.r_[0.2, 0.01, 0.02],
        'smexp': np.r_[0.2*lamb[0]*np.exp(rate/lamb[0]), rate],
        'smgauss': np.r_[1, 0.004, 0.003e-10],
        'smlorentz': np.r_[0.9, -0.006, 0.003e-10],
        'smpvoigt': np.r_[1.1, -0.001, 0.003e-10, 0.003e-10],
        'smvoigt': np.r_[1.1, -0.001, 0.003e-10, 0.003e-10],
        'smpulse_exp': np.r_[2, 3.905e-10, 0.001e-10, 0.004e-10],
        'smpulse_gauss': np.r_[1, 3.935e-10, 0.001e-10, 0.007e-10],
        'smlognorm': np.r_[amp, tau, mu, sigma],

        # double simple
        'smgauss2': np.r_[1, 0.004, 0.003e-10],
        'smlorentz2': np.r_[0.9, -0.006, 0.003e-10],
        'smpvoigt2': np.r_[1.1, -0.001, 0.003e-10, 0.003e-10],
        'smvoigt2': np.r_[1.1, -0.001, 0.003e-10, 0.003e-10],
        'smpulse_exp2': np.r_[2, 3.905e-10, 0.001e-10, 0.004e-10],
        'smpulse_gauss2': np.r_[1, 3.935e-10, 0.001e-10, 0.007e-10],
        'smlognorm2': np.r_[amp, tau, mu, sigma],

        # testing complex models
        'sm00': np.r_[
           0.2, 0.01, 0.02,
            1, 0.004, 0.003e-10,
            0.9, -0.006, 0.003e-10,
            1.1, -0.001, 0.003e-10, 0.003e-10,
        ],
        'sm01': np.r_[
            0.2*lamb[0]*np.exp(rate/lamb[0]), rate,
            1, 0.001, 0.003e-10,
            -0.001, 0.001e-10,
            1.e-12, 0.001,
        ][None, :] * np.exp(-(t[:, None] - tm)**2 / 2**2),
        'sm02': np.r_[
            0.2*lamb[0]*np.exp(rate/lamb[0]), rate,
            1, 0.001, 0.003e-10,
            -0.001,
            1e-12, 0.001, 0.003e-10, 0.001e-10,
        ],
        'sm03': np.r_[
            0.2, 0.01, 0.02,
            2, 3.905e-10, 0.001e-10, 0.004e-10,
            1, 3.935e-10, 0.001e-10, 0.007e-10,
            amp, tau, mu, sigma,
            amp, tau + 0.015e-10, mu, sigma,
        ],
    }
    return dxfree




# ###################################################
# ###################################################
#               dmodels
# ###################################################


def add_models(coll=None, models=None, lamb=_LAMB, lamb0=_LAMB0):

    # ---------------
    # check

    if coll.dobj.get('spect_model') is not None:
        return

    # --------------------------------------------------------------
    # add spectral lines just for testing automated loading of lamb0

    coll.add_spectral_line(
        key='sl00',
        ion='Ar16+',
        lamb0=3.96e-10,
    )

    lambD = lamb[-1] - lamb[0]

    # ---------------
    # dmodels

    dmodel = {
        # Testing single fit only
        'smpoly': {
            'bck0': 'poly',
        },
        'smexp': {
            'bck0': 'exp_lamb',
        },
        'smgauss': {
            'l00': 'gauss',
        },
        'smlorentz': {
            'l00': 'lorentz',
        },
        'smpvoigt': {
            'l00': 'pvoigt',
        },
        'smvoigt': {
            'l00': 'voigt',
        },
        'smpulse_exp': {
            'l00': 'pulse_exp',
        },
        'smpulse_gauss': {
            'l00': 'pulse_gauss',
        },
        'smlognorm': {
            'l00': 'lognorm',
        },

        # Testing double fits with contraints
        'smgauss2': {
            'l00': {'type': 'gauss', 'lamb0': lamb0[0]},
            'l01': {'type': 'gauss', 'lamb0': lamb0[0]},
        },
        'smlorentz2': {
            'l00': {'type': 'lorentz', 'lamb0': lamb0[0]},
            'l01': {'type': 'lorentz', 'lamb0': lamb0[0]},
        },
        'smpvoigt2': {
            'l00': {'type': 'pvoigt', 'lamb0': lamb0[0]},
            'l01': {'type': 'pvoigt', 'lamb0': lamb0[0]},
        },
        'smvoigt2': {
            'l00': {'type': 'voigt', 'lamb0': lamb0[0]},
            'l01': {'type': 'voigt', 'lamb0': lamb0[0]},
        },
        'smpulse_exp2': {
            'l00': {'type': 'pulse_exp', 'lamb0': lamb0[0]},
            'l01': {'type': 'pulse_exp', 'lamb0': lamb0[0]},
        },
        'smpulse_gauss2': {
            'l00': {'type': 'pulse_gauss', 'lamb0': lamb0[0]},
            'l01': {'type': 'pulse_gauss', 'lamb0': lamb0[0]},
        },
        'smlognorm2': {
            'l00': {'type': 'lognorm', 'lamb0': lamb0[0]},
            'l01': {'type': 'lognorm', 'lamb0': lamb0[0]},
        },

        # testing model population
        'sm-1': {
            'bck0': 'poly',
            'll0': 'gauss',
            'll1': 'gauss',
            'll2': 'lorentz',
        },
        'sm00': {
            'bck0': 'poly',
            'l00': {'type': 'gauss'},     # from spectral line
            'l08': {'type': 'lorentz'},   # from spectral line
            'l12': {'type': 'pvoigt'},    # from spectral line
        },
        'sm01': {
            'bck0': 'exp_lamb',
            'l00': {
                'type': 'gauss',
                'lamb0': 3.92e-10,
                'mz': 39.948*scpct.m_u,
            },
            'l01': {
                'type': 'lorentz',
                'lamb0': 3.95e-10,
                'mz': 39.948*scpct.m_u,
            },
            'l02': {
                'type': 'gauss',
                'lamb0': 3.97e-10,
                'mz': 39.948*scpct.m_u,
            },
        },
        'sm02': {
            'bck0': 'exp_lamb',
            'l00': {'type': 'gauss', 'lamb0': 3.92e-10},
            'l01': {'type': 'lorentz', 'lamb0': 3.95e-10},
            'l02': {'type': 'voigt', 'lamb0': 3.97e-10},
        },
        'sm03': {
            'bck0': 'poly',
            'l00': {'type': 'pulse_exp'},
            'l01': {'type': 'pulse_gauss'},
            'l02': {'type': 'lognorm'},
            'l03': {'type': 'lognorm'},
        },
    }

    dconstraints = {
        # double
        'smgauss2': {
            'g00': {'ref': 'l00_amp', 'l01_amp': [0, 0.5, 0]},
            'g01': {'ref': 'l00_sigma', 'l01_sigma': [0, 1, 0]},
            'g02': {'ref': 'l00_vccos', 'l01_vccos': [(lamb0[1] - lamb0[0])/lamb0[0], 1, 0]},
        },
        'smlorentz2': {
            'g00': {'ref': 'l00_amp', 'l01_amp': [0, 0.5, 0]},
            'g01': {'ref': 'l00_gam', 'l01_gam': [0, 1, 0]},
            'g02': {'ref': 'l00_vccos', 'l01_vccos': [(lamb0[1] - lamb0[0])/lamb0[0], 1, 0]},
        },
        'smpvoigt2': {
            'g00': {'ref': 'l00_amp', 'l01_amp': [0, 0.5, 0]},
            'g01': {'ref': 'l00_sigma', 'l01_sigma': [0, 1, 0]},
            'g02': {'ref': 'l00_gam', 'l01_gam': [0, 1, 0]},
            'g03': {'ref': 'l00_vccos', 'l01_vccos': [(lamb0[1] - lamb0[0])/lamb0[0], 1, 0]},
        },
        'smvoigt2': {
            'g00': {'ref': 'l00_amp', 'l01_amp': [0, 0.5, 0]},
            'g01': {'ref': 'l00_sigma', 'l01_sigma': [0, 1, 0]},
            'g02': {'ref': 'l00_gam', 'l01_gam': [0, 1, 0]},
            'g03': {'ref': 'l00_vccos', 'l01_vccos': [(lamb0[1] - lamb0[0])/lamb0[0], 1, 0]},
        },
        'smpulse_exp2': {
            'g00': {'ref': 'l00_amp', 'l01_amp': [0, 0.5, 0]},
            'g01': {'ref': 'l00_t_up', 'l01_t_up': [0, 1, 0]},
            'g02': {'ref': 'l00_t_down', 'l01_t_down': [0, 1, 0]},
            'g03': {'ref': 'l00_tau', 'l01_tau': [0.04e-10 / lambD, 1, 0]},
        },
        'smpulse_gauss2': {
            'g00': {'ref': 'l00_amp', 'l01_amp': [0, 0.5, 0]},
            'g01': {'ref': 'l00_t_up', 'l01_t_up': [0, 1, 0]},
            'g02': {'ref': 'l00_t_down', 'l01_t_down': [0, 1, 0]},
            'g03': {'ref': 'l00_tau', 'l01_tau': [0.04e-10 / lambD, 1, 0]},
        },
        'smlognorm2': {
            'g00': {'ref': 'l00_amp', 'l01_amp': [0, 0.5, 0]},
            'g01': {'ref': 'l00_mu', 'l01_mu': [0, 1, 0]},
            'g02': {'ref': 'l00_sigma', 'l01_sigma': [0, 1, 0]},
            'g03': {'ref': 'l00_tau', 'l01_tau': [0.04e-10 / lambD, 1, 0]},
        },

        # multi
        'sm01': {
            'g00': {'ref': 'l00_amp', 'l01_amp': [0, 1, 0]},
            'g01': {'ref': 'l00_sigma', 'l02_sigma': [0, 1, 0]},
        },
        'sm02': {
            'g00': {'ref': 'l00_amp', 'l01_amp': [0, 1, 0]},
            'g01': {'ref': 'l00_sigma', 'l01_gam': [0, 1, 0]},
        },
    }

    if models is None:
        models = sorted(dmodel.keys())

    lout = [k0 for k0 in models if k0 not in dmodel.keys()]
    if len(lout):
        lstr = [f"\t- {k0}" for k0 in lout]
        msg = "Requested models not available for tests:\n" + "\n".join(lstr)
        raise Exception(msg)

    # ---------------
    # add models
    # ---------------

    for k0 in dmodel.keys():

        if k0 not in models:
            continue

        # check err
        if k0 == 'sm-1':
            try:
                coll.add_spectral_model(
                    key='sm-1',
                    dmodel=dmodel['sm-1'],
                    dconstraints=None,
                )
                raise Exception('sucess')
            except Exception as err:
                if "For model" not in str(err):
                    raise Exception("Wrong error raised!") from err

        else:
            # no constraints
            coll.add_spectral_model(
                key=k0,
                dmodel=dmodel[k0],
                dconstraints=dconstraints.get(k0),
            )

    return


# ###################################################
# ###################################################
#               spectral model - func
# ###################################################


def get_spectral_model_func(coll=None):

    # ---------------
    # check

    if coll.dobj.get('spect_model') is None:
        add_models(coll)

    # ---------------
    # get func models

    wsm = coll._which_model
    for kmodel in coll.dobj[wsm].keys():

        for ff in ['sum', 'cost']:  # , 'details', 'jac']:

            try:
                _ = coll.get_spectral_fit_func(
                    key=kmodel,
                    func=ff,
                )

            except Exception as err:
                msg = (
                    "Could not get func for:\n"
                    f"\t- spectral model: {kmodel}\n"
                    f"\t- func: {ff}\n"
                )
                raise Exception(msg) from err

    return


# ###################################################
# ###################################################
#               spectral model - interpolate
# ###################################################


def interpolate_spectral_model(coll=None):

    # ---------------
    # check

    if coll.dobj.get('spect_model') is None:
        add_models(coll)

    # -------------
    # lamb

    lamb = np.linspace(3.9, 4, 100)*1e-10

    # ---------------
    # add model data

    wsm = coll._which_model
    lkstore = []
    for ii, kmodel in enumerate(coll.dobj[wsm].keys()):

        # interpolate
        for jj, details in enumerate([False, True]):
            for kk, store in enumerate([False, True]):

                store_key = f'interp_{kmodel}_{jj}_{kk}'
                lambi = ('lamb' if store else lamb)

                _ = coll.interpolate_spectral_model(
                    key_model=kmodel,
                    key_data=f'xfree_{kmodel}',
                    lamb=lambi,
                    # details
                    details=details,
                    # store
                    returnas=None,
                    store=store,
                    store_key=store_key,
                )

                if store:
                    lkstore.append(store_key)

    # remove stored output
    coll.remove_data(lkstore)

    return



# ###################################################
# ###################################################
#               spectral model - moments
# ###################################################


def get_spectral_model_moments(coll=None):

    # ------------
    # check

    if coll.dobj.get('spect_model') is None:
        add_models(coll)

    # ------------
    # get moments

    wsm = coll._which_model
    for kmodel in coll.dobj[wsm].keys():
        _ = coll.get_spectral_model_moments(
            kmodel,
            key_data=f"xfree_{kmodel}",
            lamb='lamb',
        )

    return


# ###################################################
# ###################################################
#               spectral model - plot
# ###################################################


def plot_spectral_model(coll=None):

    # ---------------
    # check

    if coll.dobj.get('spect_model') is None:
        add_models(coll)

    # ---------------
    # get func models

    wsm = coll._which_model
    for kmodel in coll.dobj[wsm].keys():
        pass

    return


# ###################################################
# ###################################################
#               spectral fit - add
# ###################################################


def add_fit(coll=None, key_model=None, key_data=None):

    # ---------------
    # check
    # ---------------

    add_models(coll)

    if coll.dobj.get('spect_fit') is not None:
        lk = [
            k0 for k0, v0 in coll.dobj['spect_fit'].items()
            if v0['key_data'] == key_data
            and v0['key_model'] == key_model
        ]
        if len(lk) > 0:
            return

    # ---------------
    # add 1d
    # ---------------

    mask = [None, _MASK_1D]
    domain = [
        None,
        {'lamb': [(3.96e-10, 3.97e-10)]},
        {'lamb': [
            [3.91e-10, 3.94e-10],
            (3.96e-10, 3.97e-10),
            [3.965e-10, 3.995e-10]
        ]},
    ]
    focus = [
        (None, None),
        ([[3.925e-10, 3.94e-10], [3.97e-10, 3.99e-10]], 'min'),
        ([[3.925e-10, 3.94e-10], [3.97e-10, 3.99e-10]], 'max'),
        ([[3.925e-10, 3.94e-10], [3.97e-10, 3.99e-10]], 'sum'),
    ]

    for ii, ind in enumerate(itt.product(mask, domain, focus)):

        try:
            coll.add_spectral_fit(
                key=None,
                key_model=key_model,
                key_data=key_data,
                key_sigma=None,
                key_lamb='lamb',
                # params
                dparams=None,
                dvalid={
                    'mask': ind[0],
                    'domain': ind[1],
                    'focus': ind[2][0],
                    'focus_logic': ind[2][1],
                },
            )

        except Exception as err:
            msg = (
                "Failed add_spectral_fit for 'data1d':\n"
                f"\t- ii = {ii}\n"
                f"\t- mask = {ind[0]}\n"
                f"\t- domain = {ind[1]}\n"
                + "-"*20 + "\n"
            )
            print(msg)
            raise err

    return


def add_fit_single(coll=None):

    lk = [
        ('smpoly', 'data_poly'),
        ('smexp', 'data_exp'),
        ('smgauss', 'data_gauss'),
        ('smlorentz', 'data_lorentz'),
        ('smpvoigt', 'data_pvoigt'),
        ('smpulse_exp', 'data_pulse_exp'),
        ('smpulse_gauss', 'data_pulse_gauss'),
        ('smlognorm', 'data_lognorm'),
    ]

    for (key_model, key_data) in lk:
        add_fit(coll=coll, key_model=key_model, key_data=key_data)

    return


def add_fit_double(coll=None):

    lk = [
        ('smgauss2', 'data_gauss2'),
        ('smlorentz2', 'data_lorentz2'),
        ('smpvoigt2', 'data_pvoigt2'),
        ('smpulse_exp2', 'data_pulse_exp2'),
        ('smpulse_gauss2', 'data_pulse_gauss2'),
        ('smlognorm2', 'data_lognorm2'),
    ]

    for (key_model, key_data) in lk:
        add_fit(coll=coll, key_model=key_model, key_data=key_data)

    return


def add_fit_multi(coll=None):

    lk = [
        # ('sm00', 'data1d'),
        ('sm01', 'data1d'),
    ]

    for (key_model, key_data) in lk:
        add_fit(coll=coll, key_model=key_model, key_data=key_data)

    return


# ###################################################
# ###################################################
#           plot spectral fit input validity
# ###################################################


def plot_input_validity(coll=None, key_data=None):

    # ---------------
    # check

    add_models(coll)

    wsm = coll._which_model
    for k0 in coll.dobj[wsm].keys():
        add_fit(coll, key_model=k0, key_data=key_data)

    # ---------------
    # select data

    lk = [
        k0 for k0, v0 in coll.dobj['spect_fit'].items()
        if v0['key_data'] == key_data
    ]

    # ---------------
    # plot

    for k0 in lk:
        _ = coll.plot_spectral_fit_input_validity(k0)
        plt.close('all')

    return


# ###################################################
# ###################################################
#               spectral fit - compute
# ###################################################


def compute_fit(coll=None, key_data=None, binning=None):

    # ---------------
    # select data
    # ---------------

    wsf = coll._which_fit
    lk = [
        k0 for k0, v0 in coll.dobj[wsf].items()
        if v0['key_data'] == key_data
    ]

    # ---------------
    # compute fit
    # ---------------

    for ii, k0 in enumerate(lk):

        if ii % 2 == 0:
            solver = 'scipy.least_squares'
            ftol = 1e-6
        else:
            solver = 'scipy.curve_fit'
            ftol = 1e-3

        coll.compute_spectral_fit(
            key=k0,
            strict=True,
            binning=binning,
            solver=solver,
            dsolver_options={'ftol': ftol},
            verb=1,
            timing=None,
        )

    return


def compute_fit_single(coll=None, binning=None):

    # --------------------
    # add models if needed
    # --------------------

    add_models(coll)

    # ---------------
    # add fits if needed
    # ---------------

    add_fit_single(coll)

    # --------------------
    # compute
    # --------------------

    lk = [
        'data_poly', 'data_exp',
        'data_gauss', 'data_lorentz', 'data_pvoigt',
        'data_pulse_exp', 'data_pulse_gauss', 'data_lognorm',
    ]

    for key_data in lk:
        compute_fit(coll, key_data=key_data, binning=binning)

    return


def compute_fit_double(coll=None, binning=None):

    # --------------------
    # add models if needed
    # --------------------

    add_models(coll)

    # ---------------
    # add fits if needed
    # ---------------

    add_fit_double(coll)

    # --------------------
    # compute
    # --------------------

    lk = [
        'data_gauss2', 'data_lorentz2', 'data_pvoigt2',
        'data_pulse_exp2', 'data_pulse_gauss2', 'data_lognorm2',
    ]

    for key_data in lk:
        compute_fit(coll, key_data=key_data, binning=binning)

    return


def compute_fit_multi(coll=None, key_data=None, binning=None):

    # --------------------
    # add models if needed
    # --------------------

    add_models(coll)

    # ---------------
    # add fits if needed
    # ---------------

    add_fit_multi(coll)

    # --------------------
    # compute
    # --------------------

    compute_fit(coll, key_data=key_data, binning=binning)

    return


# -*- coding: utf-8 -*-
"""
Created on Wed Jul  3 09:57:00 2024

@author: dvezinet
"""


import itertools as itt


import numpy as np
import scipy.constants as scpct
import astropy.units as asunits
import datastock as ds


from . import _class01_model_dict as _model_dict
from . import _class01_interpolate as _interpolate


#############################################
#############################################
#       main
#############################################


def main(
    coll=None,
    key=None,
    key_data=None,
    lamb=None,
    dmz=None,
    # return
    returnas=None,
):

    # ------------
    # check inputs
    # ------------

    # key_model vs key_fit
    returnas = _check(
        returnas=returnas,
    )

    # all other variables
    (
        key_model, ref_nx, ref_nf,
        key_data, key_cov, axis,
        key_lamb, lamb, ref_lamb,
        binning, details,
        _,
        _, store, store_key,
        _,
    ) = _interpolate._check(
        coll=coll,
        key_model=key,
        key_data=key_data,
        lamb=lamb,
        # others
        returnas=None,
        store=None,
        store_key=None,
    )

    # -------------------------
    # prepare model parameters
    # -------------------------

    # dconstraints
    wsm = coll._which_model
    dconstraints = coll.dobj[wsm][key_model]['dconstraints']

    # coefs
    c0 = dconstraints['c0']
    c1 = dconstraints['c1']
    c2 = dconstraints['c2']

    # param_val
    param_val = coll.get_spectral_model_variables(
        key_model,
        returnas='param_value',
        concatenate=True,
    )['param_value']

    # dind
    dind = coll.get_spectral_model_variables_dind(key_model)

    # ------------
    # mz
    # ------------

    dind = _check_mz(dmz, dind=dind)

    # ------------
    # get func
    # ------------

    func = _get_func_moments(
        c0=c0,
        c1=c1,
        c2=c2,
        dind=dind,
        param_val=param_val,
        axis=axis,
    )

    # -------------------
    # extract / compute
    # -------------------

    dout = func(
        x_free=coll.ddata[key_data]['data'],
        x_cov=None if key_cov is None else coll.ddata[key_cov]['data'],
        lamb=lamb,
    )

    # -------------
    # format output
    # -------------

    if returnas == 'dict_varnames':

        ref = tuple([rr for rr in coll.ddata[key_data]['ref'] if rr != ref_nx])

        dout = _format(
            coll=coll,
            key_model=key_model,
            key_data=key_data,
            key_lamb=key_lamb,
            din=dout,
            dind=dind,
            axis=axis,
            ref=ref,
            binning=binning,
        )

    return dout


#############################################
#############################################
#       check
#############################################


def _check(
    returnas=None,
):

    # -------------
    # returnas
    # -------------

    returnas = ds._generic_check._check_var(
        returnas, 'returnas',
        types=str,
        default='dict_varnames',
        allowed=['dict_ftypes', 'dict_varnames'],
    )

    return returnas


def _check_mz(
    dmz=None,
    dind=None,
):

    # -----------
    # trivial
    # ----------

    # add mz if user-provided
    if dmz is not None:

        for ktype in ['gauss', 'pvoigt', 'voigt']:

            if dind.get(ktype) is None:
                continue

            dind[ktype]['mz']

    return dind



#############################################
#############################################
#       moments
#############################################


def _get_func_moments(
    c0=None,
    c1=None,
    c2=None,
    dind=None,
    param_val=None,
    axis=None,
):

    # --------------
    # prepare
    # --------------

    # --------------
    # prepare
    # --------------

    def func(
        x_free=None,
        x_cov=None,
        lamb=None,
        param_val=param_val,
        c0=c0,
        c1=c1,
        c2=c2,
        dind=dind,
        axis=axis,
    ):

        # ----------
        # prepare

        lambD = lamb[-1] - lamb[0]
        lambm = 0.5*(lamb[0] + lamb[-1])

        # ----------
        # initialize

        dout = {
            k0: {} for k0 in dind.keys()
            if k0 not in ['func', 'nfunc', 'jac']
        }

        # --------------------
        # cov to std

        if x_cov is not None:
            # see _class02_compute_fit.py, line 294
            # ref_cov_axis = (axis, axis+1)
            x_std = np.sqrt(np.diagonal(x_cov, axis1=axis, axis2=axis+1))
            # diagonal always returns diag as last index
            if axis < x_cov.ndim-2:
                x_std = np.swapaxes(x_std, -1, axis)
        else:
            x_std = None

        # ----------------------------
        # get x_full from constraints

        if c0 is None:
            x_full = x_free
            if x_std is not None:
                xf_min = x_full - x_std
                xf_max = x_full + x_std

        else:

            if x_free.ndim > 1:
                shape = list(x_free.shape)
                shape[axis] = c0.size
                x_full = np.full(shape, np.nan)
                if x_std is not None:
                    xf_min = np.full(shape, np.nan)
                    xf_max = np.full(shape, np.nan)

                sli = list(shape)
                sli[axis] = slice(None)
                sli = np.array(sli)
                ich = np.array([ii for ii in range(len(shape)) if ii != axis])
                linds = [range(shape[ii]) for ii in ich]

                for ii, ind in enumerate(itt.product(*linds)):
                    sli[ich] = ind
                    slii = tuple(sli)
                    x_full[slii] = (
                        c2.dot(x_free[slii]**2) + c1.dot(x_free[slii]) + c0
                    )

                    if x_std is not None:
                        xf_min[slii] = (
                            c2.dot((x_free[slii] - x_std[slii])**2)
                            + c1.dot(x_free[slii] - x_std[slii]) + c0
                        )
                        xf_max[slii] = (
                            c2.dot((x_free[slii] + x_std[slii])**2)
                            + c1.dot(x_free[slii] + x_std[slii]) + c0
                        )

            else:
                x_full = c2.dot(x_free**2) + c1.dot(x_free) + c0
                if x_std is not None:
                    xf_min = c2.dot((x_free - x_std)**2) + c1.dot(x_free - x_std) + c0
                    xf_max = c2.dot((x_free + x_std)**2) + c1.dot(x_free + x_std) + c0

        sli = [None if ii == axis else slice(None) for ii in range(x_free.ndim)]
        extract = _get_var_extract_func(dind, axis, sli)

        # ---------------------
        # extract all variables

        for ktype, v0 in _model_dict._DMODEL.items():
            if dind.get(ktype) is not None:
                for kvar in v0['var']:
                    dout[ktype][kvar] = extract(ktype, kvar, x_full)
                    if x_std is not None:
                        dout[ktype][f"{kvar}_min"] = extract(ktype, kvar, xf_min)
                        dout[ktype][f"{kvar}_max"] = extract(ktype, kvar, xf_max)

        # ------------------
        # sum all poly

        ktype = 'poly'
        if dind.get(ktype) is not None:

            a0 = dout[ktype]['a0']
            a1 = dout[ktype]['a1']
            a2 = dout[ktype]['a2']

            # integral
            if lamb is not None:
                dout[ktype]['integ'] = (
                    a0 * (lamb[-1] - lamb[0])
                    + a1 * (lamb[-1]**2 - lamb[0]**2)/2
                    + a2 * (lamb[-1]**3 - lamb[0]**3)/3
                )

            # argmax, max
            dout[ktype]['argmax'] = np.full(a0.shape, np.nan)
            dout[ktype]['max'] = np.full(a1.shape, np.nan)
            iok = a2 != 0
            dout[ktype]['argmax'][iok] = lambm - lambD * a1[iok]/(2*a2[iok])
            dout[ktype]['max'][iok] = a0[iok] - a1[iok]**2 / (4*a2[iok])

        # --------------------
        # sum all exponentials

        ktype = 'exp_lamb'
        if dind.get(ktype) is not None:

            # physics
            rate = dout[ktype]['rate']
            dout[ktype]['Te'] = (scpct.h * scpct.c / rate) / scpct.e

            # integral
            if lamb is not None:
                amp = dout[ktype]['rate']
                dout[ktype]['integ'] = (
                    (amp / rate)
                    * (np.exp(lamb[-1] * rate) - np.exp(lamb[0] * rate))
                )

        # -----------------
        # sum all gaussians

        ktype = 'gauss'
        if dind.get(ktype) is not None:

            amp = dout[ktype]['amp']
            sigma = dout[ktype]['sigma']
            vccos = dout[ktype]['vccos']

            # argmax
            dout[ktype]['argmax'] = _get_line_argmax(
                vccos, param_val, dind, ktype, amp.shape, axis,
            )

            # integral
            dout[ktype]['integ'] = amp * sigma * np.sqrt(2 * np.pi)

            # physics
            if dind[ktype].get('mz') is not None:
                dout[ktype]['Ti'] = _get_Ti(
                    sigma,
                    param_val,
                    dind,
                    ktype,
                    sigma.shape,
                    axis,
                )

        # -------------------
        # sum all Lorentzians

        ktype = 'lorentz'
        if dind.get(ktype) is not None:

            amp = dout[ktype]['amp']
            gam = dout[ktype]['gam']
            vccos = dout[ktype]['vccos']

            # argmax
            dout[ktype]['argmax'] = _get_line_argmax(
                vccos, param_val, dind, ktype, amp.shape, axis,
            )

            # integral
            dout[ktype]['integ'] = amp * np.pi * gam

        # --------------------
        # sum all pseudo-voigt

        ktype = 'pvoigt'
        if dind.get(ktype) is not None:

            amp = dout[ktype]['amp']
            sigma = dout[ktype]['sigma']
            vccos = dout[ktype]['vccos']

            # argmax
            dout[ktype]['argmax'] = _get_line_argmax(
                vccos, param_val, dind, ktype, amp.shape, axis,
            )

            # integral
            dout[ktype]['integ'] = np.full(sigma.shape, np.nan)

            # physics
            if dind[ktype].get('mz') is not None:
                dout[ktype]['Ti'] = _get_Ti(
                    sigma,
                    param_val,
                    dind,
                    ktype,
                    sigma.shape,
                    axis,
                )

        # --------------------
        # sum all voigt

        ktype = 'voigt'
        if dind.get(ktype) is not None:

            amp = dout[ktype]['amp']
            sigma = dout[ktype]['sigma']
            vccos = dout[ktype]['vccos']

            # argmax
            dout[ktype]['argmax'] = _get_line_argmax(
                vccos, param_val, dind, ktype, amp.shape, axis,
            )

            # integral
            dout[ktype]['integ'] = amp

            # physics
            if dind[ktype].get('mz') is not None:
                dout[ktype]['Ti'] = _get_Ti(
                    sigma,
                    param_val,
                    dind,
                    ktype,
                    sigma.shape,
                    axis,
                )

        # ------------------
        # sum all pulse_exp

        ktype = 'pulse_exp'
        if dind.get(ktype) is not None:

            amp = dout[ktype]['amp']
            tau = dout[ktype]['tau']
            t_down = dout[ktype]['t_down']
            t_up = dout[ktype]['t_up']

            # integral
            dout[ktype]['integ'] = amp * (t_down - t_up)

            # prepare
            t0 = lamb[0] + lambD * tau
            dtdu = t_down - t_up
            lntdu = np.log(t_down / t_up)

            # position of max
            dout[ktype]['t0'] = t0
            dout[ktype]['argmax'] = t0 + lntdu * t_down*t_up / dtdu

            # value at max
            dout[ktype]['max'] = amp * (
                np.exp(-lntdu * t_up / dtdu)
                - np.exp(-lntdu * t_down / dtdu)
            )

        # ------------------
        # sum all pulse_gauss

        ktype = 'pulse_gauss'
        if dind.get(ktype) is not None:

            amp = dout[ktype]['amp']
            tau = dout[ktype]['tau']
            t_down = dout[ktype]['t_down']
            t_up = dout[ktype]['t_up']

            # integral
            dout[ktype]['integ'] = amp/2 * np.sqrt(np.pi) * (t_up + t_down)

            # prepare
            t0 = lamb[0] + lambD * tau

            # position of max
            dout[ktype]['t0'] = t0
            dout[ktype]['argmax'] = t0

            # value at max
            dout[ktype]['max'] = amp

        # ------------------
        # sum all lognorm

        ktype = 'lognorm'
        if dind.get(ktype) is not None:

            amp = dout[ktype]['amp']
            tau = dout[ktype]['tau']
            sigma = dout[ktype]['sigma']
            mu = dout[ktype]['mu']

            # integral
            dout[ktype]['integ'] = np.full(mu.shape, np.nan)

            # prepare
            t0 = lamb[0] + lambD * tau

            # position of max
            dout[ktype]['t0'] = t0
            dout[ktype]['argmax'] = t0 + np.exp(mu - sigma**2)

            # value at max
            dout[ktype]['max'] = amp * np.exp(0.5*sigma**2 - mu)

        return dout

    return func


# #####################################################################
# #####################################################################
#                   Mutualizing
# #####################################################################


def _get_var_extract_func(dind, axis, sli):
    def func(ktype, kvar, x_full, dind=dind, axis=axis, sli=sli):
        sli[axis] = dind[ktype][kvar]['ind']
        return x_full[tuple(sli)]
    return func


def _get_line_argmax(vccos, param_val, dind, ktype, shape, axis):

    # lamb0
    lamb0 = param_val[dind[ktype]['lamb0']]

    # reshape lamb0
    reshape = [1 for ii in shape]
    reshape[axis] = lamb0.size
    lamb0 = lamb0.reshape(tuple(reshape))

    return lamb0 * (1 + vccos)


def _get_Ti(sigma, param_val, dind, ktype, shape, axis):

    # lamb0, mz
    lamb0 = param_val[dind[ktype]['lamb0']]
    mz = param_val[dind[ktype]['mz']]

    # reshape lamb0 and mz
    reshape = [1 for ii in shape]
    reshape[axis] = lamb0.size
    reshape = tuple(reshape)
    lamb0 = lamb0.reshape(reshape)
    mz = mz.reshape(reshape)

    return (sigma / lamb0)**2 * mz * scpct.c**2 / scpct.e


#############################################
#############################################
#       format
#############################################


def _format(
    coll=None,
    key_model=None,
    key_data=None,
    key_lamb=None,
    din=None,
    dind=None,
    axis=None,
    ref=None,
    binning=None,
):

    # ---------------
    # prepare
    # ---------------

    dout = {}
    wsm = coll._which_model
    sli = [slice(None) for ii in range(len(ref)+1)]

    # -------------
    # units
    # -------------

    # data
    try:
        units_data = asunits.Unit(coll.ddata[key_data]['units'])
    except Exception as err:
        units_data = coll.ddata[key_data]['units']

    # lamb
    try:
        units_lamb = asunits.Unit(coll.ddata[key_lamb]['units'])
    except Exception as err:
        units_lamb = coll.ddata[key_lamb]['units']

    # --------------
    # loop
    # --------------

    for ktype, vtype in din.items():

        lkfunc = [
            kfunc for kfunc in coll.dobj[wsm][key_model]['keys']
            if coll.dobj[wsm][key_model]['dmodel'][kfunc]['type'] == ktype
        ]

        for kvar in vtype.keys():

            for ii, kfunc in enumerate(lkfunc):

                key = f"{kfunc}_{kvar}"
                sli[axis] = ii

                # get units
                units = _units(
                    coll, ktype, kvar,
                    units_data=units_data,
                    units_lamb=units_lamb,
                    binning=binning,
                )

                # store
                dout[key] = {
                    'data': din[ktype][kvar][tuple(sli)],
                    'ref': ref,
                    'units': units,
                }

    return dout


#############################################
#############################################
#       units
#############################################


def _units(coll, ktype, kvar, units_data=None, units_lamb=None, binning=None):

    if kvar.endswith('_min') or kvar.endswith('_max'):
        kvar = kvar[:-4]
    units = None

    # ------------
    # trivial
    # ------------

    if kvar == 'integ':
        if binning is False:
            units = _try_units_mult(units_data, units_lamb)
        else:
            units = units_data

    elif kvar == 'max':
        if binning is False:
            units = units_data
        else:
            units = _try_units_divide(units_data, units_lamb)

    elif kvar == 'argmax':
        units = units_lamb

    # ----------------------------------
    # non-trivial but common to several
    # ----------------------------------

    elif kvar == 'Ti':

        # assuming mz is provided in kg
        units = 'eV'

    # ----------------
    # specific to each
    # ----------------
    else:

        if ktype == 'poly':
            if binning is False:
                units = units_data
            else:
                units = _try_units_divide(units_data, units_lamb)

        elif ktype == 'exp_lamb':

            if kvar == 'amp':
                if binning is False:
                    units = _try_units_mult(units_data, units_lamb)
                else:
                    units = units_data

            elif kvar == 'rate':
                units = units_lamb

            elif kvar == 'Te':
                if units_lamb == 'm':
                    units = 'eV'
                else:
                    units = asunits('eV') * (asunits.Units('m') / units_lamb)

        elif ktype in ['gauss', 'lorentz', 'pvoigt', 'voigt']:

            if kvar == 'amp':
                if binning is False:
                    units = units_data
                else:
                    units = _try_units_divide(units_data, units_lamb)

            elif kvar == 'sigma':
                units = units_lamb

            elif kvar == 'gam':
                units = units_lamb

            elif kvar == 'vccos':
                units = ''

        elif ktype in ['pulse_exp', 'pulse_gauss']:

            if kvar == 'amp':
                if binning is False:
                    units = units_data
                else:
                    units = _try_units_divide(units_data, units_lamb)

            elif kvar == 'tau':
                units = ''

            elif kvar == 't0':
                units = units_lamb

            elif kvar == 't_up':
                units = units_lamb

            elif kvar == 't_down':
                units = units_lamb

        elif ktype == 'lognorm':

            if kvar == 'amp':
                if binning is False:
                    units = _try_units_mult(units_data, units_lamb)
                else:
                    units = units_data

            elif kvar == 'tau':
                units = ''

            elif kvar == 't0':
                units = units_lamb

            elif kvar == 'mu':
                units = 'TBD'

            elif kvar == 'sigma':
                units = ''

    # --------------
    # safety check
    # --------------

    if units is None:
        msg = f"units not reckognized for {ktype} {kvar}"
        raise Exception(msg)

    try:
        return asunits.Unit(units)
    except Exception as err:
        return units


# ###########################
# ###########################
#      try units
# ###########################


def _try_units_mult(u0, u1):
    try:
        units = u0 * u1
    except Exception as err:
        units = f"{u0} x {u1}"
    return units


def _try_units_divide(u0, u1):
    try:
        units = u0 / u1
    except Exception as err:
        units = f"{u0} / {u1}"
    return units


# #!/usr/bin/python3
# -*- coding: utf-8 -*-


import itertools as itt


import numpy as np
import scipy.constants as scpct
import datastock as ds


# local
from . import _class01_model_dict as _model_dict


#############################################
#############################################
#       MODEL CHECK
#############################################


def _dmodel(
    coll=None,
    key=None,
    dmodel=None,
):

    # ----------
    # key
    # ----------

    wsm = coll._which_model
    key = ds._generic_check._obj_key(
        d0=coll.dobj.get(wsm, {}),
        short='sm',
        key=key,
        ndigits=2,
    )

    # --------------
    # check dmodel
    # --------------

    dmodel = _check_dmodel(
        coll=coll,
        key=key,
        dmodel=dmodel,
    )

    # --------------
    # store
    # --------------

    # add ref_nfun
    knfunc = f"nf_{key}"
    if knfunc not in coll.dref.keys():
        coll.add_ref(knfunc, size=len(dmodel))
    elif coll.dref[knfunc]['size'] != len(dmodel):
        msg = (
            f"add_model('{key}') requires adding ref knfunc = '{knfunc}'"
            " but it already exists and doesn't have the correct value!\n"
            f"\t- existing: {coll.dref[knfunc]['size']}\n"
            f"\t- requested: {len(dmodel)}\n"
        )
        raise Exception(msg)

    # dmodel
    dobj = {
        wsm: {
            key: {
                'keys': sorted(dmodel.keys()),
                'ref_nx': None,
                'ref_nf': knfunc,
                'dmodel': dmodel,
                'dconstraints': None,
            },
        },
    }

    coll.update(dobj=dobj)

    return


#############################################
#############################################
#       check dmodel
#############################################


def _dmodel_err(key, dmodel):

    # prepare list of str
    lstr = []
    for ii, (k0, v0) in enumerate(_model_dict._DMODEL.items()):
        if v0.get('param') is None:
            stri = f"\t- 'f{ii}': '{k0}'"
        else:
            lpar = v0['param']
            pstr = ", ".join([f"'{tpar[0]}': {tpar[1]}" for tpar in lpar])
            stri = f"\t- 'f{ii}': " + "{" + f"'type': '{k0}', {pstr}" + "}"

        if k0 == 'poly':
            lstr.append("\t# background-oriented")
        elif k0 == 'gauss':
            lstr.append("\t# spectral lines-oriented")
        elif k0 == 'pulse_exp':
            lstr.append("\t# pulse-oriented")
        lstr.append(stri)

    # Provided
    if isinstance(dmodel, dict):
        prov = "\n".join([f"\t'{k0}': {v0}," for k0, v0 in dmodel.items()])
        prov = "{\n" + prov + "\n}"
    else:
        prov = str(dmodel)

    # concatenate msg
    return (
        f"For model '{key}' dmodel must be a dict of the form:\n"
         + "\n".join(lstr)
         + f"\n\nProvided:\n{prov}"
    )


def _check_dmodel(
    coll=None,
    key=None,
    dmodel=None,
):

    # -------------
    # model
    # -------------

    if isinstance(dmodel, str):
        dmodel = [dmodel]

    if isinstance(dmodel, (tuple, list)):
        dmodel = {ii: mm for ii, mm in enumerate(dmodel)}

    if not isinstance(dmodel, dict):
        raise Exception(_dmodel_err(key, dmodel))

    # prepare for extracting lamb0
    wsl = coll._which_lines

    # ------------
    # check dict
    # ------------

    dmod2 = {}
    dout = {}
    ibck, il = 0, 0
    for k0, v0 in dmodel.items():

        # -----------------
        # check str vs dict

        if isinstance(v0, dict):
            if isinstance(v0.get('type'), str):
                typ = v0['type']
            else:
                dout[k0] = v0
                continue

        elif isinstance(v0, str):
            typ = v0

        else:
            dout[k0] = v0
            continue

        # ----------
        # check type

        if typ not in _model_dict._DMODEL.keys():
            dout[k0] = v0
            continue

        # ----------
        # check key

        if isinstance(k0, int):
            if typ in ['poly', 'exp']:
                k1 = f'bck{ibck}'
                ibck += 1
            else:
                k1 = f"l{il}"
                il += 1
        else:
            k1 = k0

        # ---------------------------
        # check parameter (if needed)

        haspar = _model_dict._DMODEL[typ].get('param') is not None
        if haspar is True:

            lpar = _model_dict._DMODEL[typ]['param']

            # loop on parameters
            dpar = {}
            for tpar in lpar:

                # provided
                c0 = (
                    isinstance(v0, dict)
                    and isinstance(v0.get(tpar[0]), tpar[1])
                )
                if c0:
                    dpar[tpar[0]] = v0[tpar[0]]

                elif tpar[0] in ('lamb0', 'mz'):

                    # check if lamb0 can be extracted from existing lines
                    c1 = (
                        typ in ['gauss', 'lorentz', 'pvoigt', 'voigt']
                        and k1 in coll.dobj.get(wsl, {}).keys()
                    )
                    if c1 and tpar[0] == 'lamb0':
                        dpar[tpar[0]] = coll.dobj[wsl][k1]['lamb0']

                    elif c1 and tpar[0] == 'mz':
                        kion = coll.dobj[wsl][k1]['ion']
                        dpar[tpar[0]] = coll.dobj['ion'][kion]['A'] * scpct.m_u

                    elif len(tpar) == 3:
                        dpar[tpar[0]] = tpar[2]
                    else:
                        dout[k0] = v0
                        continue

                elif len(tpar) == 3:
                    dpar[tpar[0]] = tpar[2]

                else:
                    dout[k0] = v0
                    continue

        # ----------------
        # assemble

        dmod2[k1] = {
            'type': typ,
            'var': _model_dict._DMODEL[typ]['var'],
        }

        # add parameter
        if haspar is True:
            dmod2[k1]['param'] = dpar

    # ---------------
    # raise error
    # ---------------

    if len(dout) > 0:
        raise Exception(_dmodel_err(key, dout))

    return dmod2


#############################################
#############################################
#       Get variables
#############################################


def _get_var(
    coll=None,
    key=None,
    concatenate=None,
    returnas=None,
):

    # --------------
    # check key
    # -------------

    # key
    wsm = coll._which_model
    lok = list(coll.dobj.get(wsm, {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )

    # keys and dmodel
    keys = coll.dobj[wsm][key]['keys']
    dmodel = coll.dobj[wsm][key]['dmodel']

    # returnas
    if isinstance(returnas, str):
        returnas = [returnas]
    returnas = ds._generic_check._check_var_iter(
        returnas, 'returnas',
        types=(list, tuple),
        types_iter=str,
        default=['all', 'param'],
        allowed=['all', 'free', 'tied', 'param_key', 'param_value'],
    )

    # concatenate
    concatenate = ds._generic_check._check_var(
        concatenate, 'concatenate',
        types=bool,
        default=True,
    )

    # -------------
    # get lvar
    # -------------

    dout = {}

    # ---------------
    # all variables

    if 'all' in returnas:
        dout['all'] = [
            [f"{k0}_{k1}" for k1 in dmodel[k0]['var']]
            for k0 in keys
        ]

    # -----------------------
    # free or tied variables

    if 'free' in returnas or 'tied' in returnas:
        dconstraints = coll.dobj[wsm][key]['dconstraints']
        lref = [v0['ref'] for v0 in dconstraints['dconst'].values()]

        # lvar
        if 'free' in returnas:
            dout['free'] = [
                [
                    f"{k0}_{k1}" for k1 in dmodel[k0]['var']
                    if f"{k0}_{k1}" in lref
                ]
                for k0 in keys
            ]

        if 'tied' in returnas:
            dout['tied'] = [
                [
                    f"{k0}_{k1}" for k1 in dmodel[k0]['var']
                    if f"{k0}_{k1}" not in lref
                ]
                for k0 in keys
            ]

    # ---------------
    # parameters

    if 'param_key' in returnas:
        dout['param_key'] = [
            [
                f"{k0}_{tpar[0]}"
                for tpar in _model_dict._DMODEL[dmodel[k0]['type']]['param']
            ]
            for k0 in keys if dmodel[k0].get('param') is not None
        ]

    if 'param_value' in returnas:
        dout['param_value'] = [
            [
                dmodel[k0]['param'][tpar[0]]
                for tpar in _model_dict._DMODEL[dmodel[k0]['type']]['param']
            ]
            for k0 in keys if dmodel[k0].get('param') is not None
        ]

    # ----------------
    # concatenate
    # ----------------

    if concatenate is True:
        for k0, v0 in dout.items():
            dout[k0] = list(itt.chain.from_iterable(v0))

            if k0 == 'param_value':
                dout[k0] = np.array(dout[k0])

    # ----------------
    # return
    # ----------------

    return dout


#############################################
#############################################
#       Get variables dind
#############################################


def _get_var_dind(
    coll=None,
    key=None,
):

    # --------------
    # check key
    # -------------

    # key
    wsm = coll._which_model
    lok = list(coll.dobj.get(wsm, {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )

    # keys and dmodel
    keys = coll.dobj[wsm][key]['keys']
    dmodel = coll.dobj[wsm][key]['dmodel']

    # -------------
    # get lvar and param
    # -------------

    dout = coll.get_spectral_model_variables(
        key,
        returnas=['all', 'free', 'param_key', 'param_value'],
        concatenate=True,
    )
    x_all = dout['all']
    x_free = dout['free']
    param_key = dout['param_key']

    # ---------------
    # derive dind
    # ---------------

    types = sorted(set([v0['type'] for v0 in dmodel.values()]))

    dind = {}
    for ktype in types:

        # list functions with corresponding model type
        lf = [k0 for k0 in keys if dmodel[k0]['type'] == ktype]

        # populate
        dind[ktype] = {
            k1: {
                'ind': np.array([x_all.index(f"{kf}_{k1}") for kf in lf]),
                'keys': [f"{kf}_{k1}" for kf in lf],
            }
            for k1 in dmodel[lf[0]]['var']
        }

        # add param
        if dmodel[lf[0]].get('param') is not None:
            for kpar in dmodel[lf[0]]['param'].keys():
                dind[ktype][kpar] = [
                    param_key.index(f"{kf}_{kpar}")
                    for kf in lf
                ]

    # ---------------
    # safety checks
    # ---------------

    # aggregate all variables
    lvar = tuple(itt.chain.from_iterable([
        list(itt.chain.from_iterable([
            v1['keys']
            for k1, v1 in vtype.items()
            if not isinstance(v1, list)
        ]))
        for ktype, vtype in dind.items()
    ]))

    # check all indices are unique
    lind = tuple(itt.chain.from_iterable([
        list(itt.chain.from_iterable([
            v1['ind']
            for k1, v1 in vtype.items()
            if not isinstance(v1, list)
        ]))
        for ktype, vtype in dind.items()
    ]))

    # check all variable are represented
    nn = len(x_all)
    c0 = (
        (tuple(sorted(lvar)) == tuple(sorted(x_all)))
        and np.allclose(sorted(lind), np.arange(nn))
        and (tuple([lvar[ii] for ii in np.argsort(lind)]) == tuple(x_all))
    )
    if not c0:
        msg = (
            "dind corrupted!\n"
            f"\t- x_all: {x_all}\n"
            f"\t- lvar: {lvar}\n"
            f"\t- lind: {lind}\n"
            f"\ndind:\n{dind}\n"
        )
        raise Exception(msg)

    # ----------------
    # add func
    # ----------------

    dind['func'] = {}
    for ktype in types:

        # list functions with corresponding model type
        lf = [k0 for k0 in keys if dmodel[k0]['type'] == ktype]

        # get indices
        ind = [keys.index(ff) for ff in lf]

        # store
        dind['func'][ktype] = {
            'keys': lf,
            'ind': np.array(ind, dtype=int),
        }

    #-------------------------
    # add total number of func

    dind['nfunc'] = len(keys)

    # ----------------
    # add jac
    # ----------------

    dind['jac'] = {ktype: {} for ktype in types}
    for ktype in types:
        lf = [k0 for k0 in keys if dmodel[k0]['type'] == ktype]
        for kvar in _model_dict._DMODEL[ktype]['var']:
            lvar = [f"{kf}_{kvar}" for kf in lf]
            lind = [ii for ii, vv in enumerate(lvar) if vv in x_free]

            inds = np.array(
                [x_free.index(lvar[ii]) for ii in lind],
                dtype=int,
            )

            if inds.size > 0:
                dind['jac'][ktype][kvar] = {
                    'val': inds,
                    'var': np.array(lind, dtype=int),
                }

    # ---------------
    # safety checks
    # ---------------

    lind = sorted(itt.chain.from_iterable([
        list(itt.chain.from_iterable([v1['val'].tolist() for v1 in v0.values()]))
        for v0 in dind['jac'].values()
    ]))
    if not np.allclose(lind, np.arange(len(x_free))):
        msg = (
            "Something wrong with dind['jac'] !\n"
            f"\t- lind = {lind}\n"
            f"\t- dind['jac']:\n{dind['jac']}\n"
        )
        raise Exception(msg)

    return dind